#!/usr/bin/python
# filename: regions.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import re
import sys
import traceback

from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from abtools import log



# Both IMGT_REGION_END_POSITIONS_AA and IMGT_REGION_END_POSITIONS_NT use the actual
# IMGT end positions, thus they're not suitable for slicing (because Python's slicing
# end point is exclusive). To use these end points in a slice, you need to add 1.
IMGT_REGION_END_POSITIONS_AA = {'FR1': 26, 'CDR1': 38, 'FR2': 55, 'CDR2': 65,
                                'FR3': 104, 'CDR3': 117, 'FR4': 129}

IMGT_REGION_END_POSITIONS_NT = {'FR1': 78, 'CDR1': 114, 'FR2': 165, 'CDR2': 195,
                                'FR3': 312, 'CDR3': 351, 'FR4': 387}

IMGT_REGION_START_POSITIONS_AA = {'FR1': 1, 'CDR1': 27, 'FR2': 39, 'CDR2': 56,
                                  'FR3': 66, 'CDR3': 105, 'FR4': 118}

IMGT_REGION_START_POSITIONS_NT = {'FR1': 1, 'CDR1': 79, 'FR2': 115, 'CDR2': 166,
                                  'FR3': 196, 'CDR3': 313, 'FR4': 352}



# def regions(antibody):
#     # V-gene regions
#     antibody.v.regions = VariableRegions(antibody)
#     # J-gene regions
#     antibody.j.regions = JoiningRegions(antibody)


def get_variable_regions(antibody):
    return VariableRegions(antibody)


def get_joining_regions(antibody):
    return JoiningRegions(antibody)


class BaseRegions(object):
    """Base class for Variable and Joining gene regions"""
    def __init__(self, antibody):
        super(BaseRegions, self).__init__()
        self.antibody = antibody

        self._raw_nt_positions = None
        self._nt_seqs = None
        self._aa_seqs = None


    @property
    def raw_nt_positions(self):
        '''
        A `dict`, with region names as keys. Values are `list`s containing the start/end points
        (in raw numbering, corresponding to the oriented_input). If raw start/end points
        couldn't be identified, they are `None`. So if an entire region isn't present (either due
        to deletion or query truncation), the value will be `[None, None]`.

        Note that the end points are designed for slicing, so the end point is actually 1 position
        beyond the actual end point. This allows `ab.oriented_input[start:end]` to work. This also
        means that the end point of FR1 should be identical to the start point of CDR1.
        '''
        if self._raw_nt_positions is None:
            pos = {}
            for region in self.region_names:
                start = self._raw_region_start_position_nt(region)
                end = self._raw_region_end_position_nt(region)
                pos[region] = [start, end]
            self._raw_nt_positions = pos
        return self._raw_nt_positions


    @property
    def nt_seqs(self):
        if self._nt_seqs is None:
            nt_seqs = {}
            for region in self.region_names:
                start, end = self.raw_nt_positions[region]
                if all([start is None, end is None]):
                    nt_seqs[region] = ''
                    continue
                if start is not None and end is None:
                    err = '{} contains a start position but not an end position'.format(region)
                    self.antibody.log('REGION NT SEQUENCE IDENTIFICATION ERROR:', err)
                elif start > end:
                    err = '{} contains a start position {} that is greater than the end position{}'.format(region, start, end)
                    self.antibody.log('REGION NT SEQUENCE IDENTIFICATION ERROR:', err)
                else:
                    # if start is None, that means the region begins before the query sequence does. But we
                    # don't necessarily want to start the region sequence at the beginning of the query,
                    # we want to start the region sequence at the point that the query aligns to the germline.
                    if start is None:
                        start = self.segment.query_start
                    nt_seq = self.antibody.oriented_input[start:end]
                    nt_seqs[region] = nt_seq
            self._nt_seqs = nt_seqs
        return self._nt_seqs


    @property
    def aa_seqs(self):
        if self._aa_seqs is None:
            to_translate = {}
            aa_seqs = {}
            first_region = self._get_first_region()
            for i, region in enumerate(self.region_names):
                # in the case of region-spanning indels, we need to process multiple regions
                # simultaneously. The following prevents us from reprocessing a region
                # that contains the second half of a region-spanning deletion.
                if region in to_translate.keys():
                    continue
                # if we're looking at the first region in the query sequence, we need to make
                # sure that we're in the correct reading frame.
                if region == first_region and self.segment.gene_type == 'V':
                    region_nt = self.nt_seqs[region][self.antibody.v_rf_offset:]
                else:
                    region_nt = self.nt_seqs[region]
                # check to make sure the region_nt sequence is a multiple of three. If it's not,
                # there are two possibilities:
                # 1) a region-spanning deletion that deletes, for example, 1nt from this region and 2nt
                #    from the following region. In this case, both this region and the next would be out
                #    of frame and we need to fix that.
                # 2) a frameshift indel within this region that's longer than a single codon (since we
                #    don't automatically fix those).
                if len(region_nt) % 3 != 0:
                    # if we're looking at the last region, we obviously can't look to the next
                    # region for the other portion of a region-spanning deletion, so we'll just
                    # try to translate the region as is (as it's either truncated at the 3' end or
                    # we have a frameshift indel that's longer than a single codon which we don't want
                    # to automatically fix).
                    if i == len(self.region_names) - 1:
                        to_translate[region] = self.nt_seqs[region]
                    else:
                        # find the next region that has sequence data. A little convoluted, but this
                        # accounts for the rare case of a large region-spanning deletion that actually
                        # deletes an entire downstream region (for example, a deletion that starts at the end
                        # of FR2, deletes the entire CDR2, and ends near the start of FR3).
                        next_regions = self.region_names[i + 1:]
                        reg_iter = iter(next_regions)
                        region2 = reg_iter.next()
                        while len(self.nt_seqs[region2].strip('-')) == 0:
                            to_translate[region2] = ''
                            region2 = reg_iter.next()
                        # if the current region and the next region with sequence data combine to form an in-frame
                        # sequence, we can fix the region-spanning deletion so that both regions are translated
                        # correctly.
                        if len(region_nt + self.nt_seqs[region2]) % 3 == 0:
                            region_seq, region2_seq = self._fix_region_spanning_indel(region_nt,
                                                                                      self.nt_seqs[region2])
                            to_translate[region] = region_seq
                            to_translate[region2] = region2_seq
                        # if the combined regions are still out of frame, the current region likely contains a
                        # frameshift indel that shouldn't be automatically fixed (it's longer than a single codon).
                        else:
                            to_translate[region] = region_nt
                else:
                    to_translate[region] = region_nt
            # translate the sequences for each region
            for region, nt_seq in to_translate.items():
                aa_seq = str(Seq(nt_seq, generic_dna).translate())
                aa_seqs[region] = aa_seq
            self._aa_seqs = aa_seqs
        return self._aa_seqs


    def _raw_region_start_position_nt(self, region):
        # get the raw start position of the region
        imgt_start = IMGT_REGION_START_POSITIONS_NT[region]
        # if there isn't a direct equivalent to the IMGT region start position,
        # that could be because the qyuery sequence is truncated (and the start of the
        # region isn't present in the query sequence) or it could be because a deletion
        # removed the region start position.
        # First, let's check to see if there's any query sequence prior to the start of
        # the region.
        if imgt_start > max(self.segment._imgt_position_from_raw.values()):
            return None
        # If there's sequence prior to the start of the region but the region start position
        # isn't present, we must have a deletion that removed the region start position. In
        # this case, we just scan forward in the sequence to find the earliest non-deleted
        # position in the region and use that as the region start position
        while self.segment.get_raw_position_from_imgt(imgt_start) is None:
            # we don't want to keep going past the end of the region so only keep looking
            # for the end of the deletion until we reach the start of the next region
            if imgt_start >= IMGT_REGION_END_POSITIONS_NT[region]:
                break
            imgt_start += 1
        return self.segment.get_raw_position_from_imgt(imgt_start)



    def _raw_region_end_position_nt(self, region):
        imgt_end = IMGT_REGION_END_POSITIONS_NT[region]
        # if the end position of the region just happens to be part of a deletion,
        # or if the query sequence doesn't contain the region, then there isn't a
        # corresponding raw sequence position.
        # First, check to see if the query sequence begins after the end of the region
        if imgt_end < min(self.segment._imgt_position_from_raw.values()):
            return None
        # Now we need to try to find the start of a deletion that includes the region
        # end position
        while self.segment.get_raw_position_from_imgt(imgt_end) is None:
            imgt_end -= 1
            # we don't want to move all the way back into a different region,
            # so only keep looking for the start of the deletion until we
            # reach the start of the region
            if imgt_end < IMGT_REGION_START_POSITIONS_NT[region]:
                break
        return self.segment.get_raw_position_from_imgt(imgt_end) + 1


    def _get_first_region(self):
        for region in self.region_names:
            if self.raw_nt_positions[region][1] is not None:
                return region


    def _fix_region_spanning_indel(self, r1_nt, r2_nt):
        # if the first region contains the majority of the split codon,
        # we add the complete codon to the first region
        if len(r1_nt) % 3 == 2:
            return r1_nt + r2_nt[0], r2_nt[1:]
        # if the second region contains the majority of the split codon,
        # we add the complete codon to the second region
        elif len(r1_nt) % 3 == 1:
            return r1_nt[:-1], r1_nt[-1] + r2_nt
        return r1_nt, r2_nt


class VariableRegions(BaseRegions):
    """docstring for VariableRegions"""
    def __init__(self, antibody):
        super(VariableRegions, self).__init__(antibody)
        self.segment = antibody.v
        self.region_names = ['FR1', 'CDR1', 'FR2', 'CDR2', 'FR3']


class JoiningRegions(BaseRegions):
    """docstring for JoiningRegions"""
    def __init__(self, antibody):
        super(JoiningRegions, self).__init__(antibody)
        self.segment = antibody.j
        self.region_names = ['FR4']



#!/usr/bin/python
# filename: mutations.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import math
import re
import traceback

from abtools import log
from abtools.utils.codons import codon_lookup as codons

from .regions import IMGT_REGION_START_POSITIONS_AA, IMGT_REGION_END_POSITIONS_AA



def nt_mutations(antibody):
    try:
        all_mutations = Mutations()
        for segment in [antibody.v, antibody.j]:
            if segment is None:
                continue
            mutations = Mutations()
            for i, (q, g) in enumerate(zip(segment.query_alignment, segment.germline_alignment)):
                if q != g:
                    raw_pos = i + segment.query_start
                    imgt_pos = segment.get_imgt_position_from_raw(raw_pos)
                    # we don't want to count indels as mutations
                    if any([q == '-', g == '-', imgt_pos is None]):
                        continue
                    imgt_codon = int(math.ceil(imgt_pos / 3.0))
                    if segment.gene_type == 'J':
                        imgt_pos = segment.correct_imgt_nt_position_from_imgt(imgt_pos)
                    mutations.add(Mutation(g, q, raw_pos, imgt_pos, imgt_codon))
            segment.nt_mutations = mutations
            segment.nt_identity = 100. - (100. * mutations.count / min(len(segment.query_alignment), len(segment.germline_alignment)))
            all_mutations.add_many(mutations.mutations)
        antibody.log('')
        antibody.log('NT MUTATIONS')
        antibody.log('------------')
        antibody.log('FR1:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR1')]))
        antibody.log('CDR1:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('CDR1')]))
        antibody.log('FR2:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR2')]))
        antibody.log('CDR2:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('CDR2')]))
        antibody.log('FR3:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR3')]))
        antibody.log('CDR3:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('CDR3')]))
        antibody.log('FR4:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR4')]))
        return all_mutations
    except:
        antibody.exception('NT MUTATIONS', traceback.format_exc())
        return Mutations()


def aa_mutations(antibody):
    try:
        all_mutations = Mutations()
        for segment in [antibody.v, antibody.j]:
            mutations = Mutations()
            for i, (q, g) in enumerate(zip(segment.query_alignment, segment.germline_alignment)):
                raw_nt_pos = i + segment.query_start
                imgt_nt_pos = segment.get_imgt_position_from_raw(raw_nt_pos)
                if imgt_nt_pos is None:
                    continue
                if imgt_nt_pos % 3 != 1:
                    continue
                if any([q == '-', g == '-']):
                    continue
                q_codon_end_pos = i + 3
                if q_codon_end_pos >= len(segment.query_alignment):
                    continue
                q_codon = segment.query_alignment[i:q_codon_end_pos]
                while len(q_codon.replace('-', '')) < 3:
                    q_codon_end_pos += 1
                    q_codon = segment.query_alignment[i:q_codon_end_pos]
                    if q_codon_end_pos >= len(segment.query_alignment):
                        break
                if len(q_codon.replace('-', '')) < 3:
                    continue
                g_codon_end_pos = i + 3
                g_codon = segment.germline_alignment[i:g_codon_end_pos]
                while len(g_codon.replace('-', '')) < 3:
                    g_codon_end_pos += 1
                    g_codon = segment.germline_alignment[i:g_codon_end_pos]
                q_aa = codons.get(q_codon.replace('-', ''), 'X')
                g_aa = codons.get(g_codon.replace('-', ''), 'X')
                if q_aa != g_aa:
                    imgt_aa_pos = int(math.ceil(imgt_nt_pos / 3.0))
                    if segment.gene_type == 'J':
                        imgt_aa_pos = segment.correct_imgt_aa_position_from_imgt(imgt_aa_pos)
                    mutations.add(Mutation(g_aa, q_aa, None, imgt_aa_pos, imgt_aa_pos))
            segment.aa_mutations = mutations
            segment.aa_identity = 100. - (100. * mutations.count / len(segment.aa_sequence))
            all_mutations.add_many(mutations.mutations)
        antibody.log('')
        antibody.log('AA MUTATIONS')
        antibody.log('------------')
        antibody.log('FR1:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR1')]))
        antibody.log('CDR1:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('CDR1')]))
        antibody.log('FR2:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR2')]))
        antibody.log('CDR2:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('CDR2')]))
        antibody.log('FR3:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR3')]))
        antibody.log('CDR3:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('CDR3')]))
        antibody.log('FR4:', ', '.join([m.abstar_formatted for m in all_mutations.in_region('FR4')]))
        return all_mutations
    except:
        antibody.exception('AA MUTATIONS', traceback.format_exc())
        return Mutations()


def _get_joining_imgt_mutation_position(codon_num, j):
    pass



class Mutation(object):
    """docstring for Mutation"""
    def __init__(self, was, now, raw_position, imgt_position, imgt_codon):
        super(Mutation, self).__init__()
        self.was = was
        self.now = now
        self.raw_position = raw_position
        self.imgt_position = imgt_position
        self.imgt_codon = imgt_codon


    @property
    def imgt_formatted(self):
        return '{}{}>{}'.format(self.was.lower(),
                                self.imgt_position,
                                self.now.lower())


    @property
    def abstar_formatted(self):
        return '{}:{}>{}'.format(self.imgt_position,
                                 self.was.upper(),
                                 self.now.upper())


    @property
    def json_formatted(self):
        return {'was': self.was,
                'is': self.now,
                'raw_position': self.raw_position,
                'position': self.imgt_position,
                'codon': self.imgt_codon}




class Mutations(object):
    """docstring for Mutations"""
    def __init__(self, mutations=None):
        super(Mutations, self).__init__()
        self.mutations = mutations if mutations is not None else []

    def __iter__(self):
        return (m for m in self.mutations)


    @property
    def count(self):
        return len(self.mutations)


    def add(self, mutation):
        '''
        Add a single mutation.

        Args:

            mutation (Mutation): the Mutation object to be added
        '''
        self.mutations.append(mutation)


    def add_many(self, mutations):
        '''
        Adds multiple mutations.

        Args:

            mutation (list): an iterable of Mutation objects to be added
        '''
        self.mutations += mutations


    def in_region(self, region):
        region_mutations = []
        start = IMGT_REGION_START_POSITIONS_AA[region]
        end = IMGT_REGION_END_POSITIONS_AA[region]
        for mut in self.mutations:
            if all([mut.imgt_codon >= start, mut.imgt_codon <= end]):
                region_mutations.append(mut)
        return region_mutations


    def in_region_count(self, region):
        return len(self.in_region(region))



#!/usr/bin/env python
# filename: abstar.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function, unicode_literals

from argparse import ArgumentParser
from glob import glob
import gzip
import logging
from multiprocessing import Pool
import os
import re
from subprocess import Popen, PIPE
import sys
import tempfile
import time
import traceback
import warnings

from Bio import SeqIO

from abtools import log
from abtools.sequence import Sequence

from .antibody import Antibody
from ..assigners.assigner import BaseAssigner
from ..assigners.registry import ASSIGNERS
# from ..utils import output
from ..utils.output import get_abstar_result, get_output, write_output, get_header
from ..utils.queue.celery import celery


# ASSIGNERS = {cls.__name__.lower(): cls for cls in vars()['BaseAssigner'].__subclasses__()}


#####################################################################
#
#                             ARGUMENTS
#
#####################################################################


def parse_arguments(print_help=False):
    parser = ArgumentParser("Performs germline assignment and other relevant annotation on antibody sequence data from NGS platforms.")
    parser.add_argument('-p', '--project', dest='project_dir', default=None,
                        help="The data directory, where files will be downloaded (or have previously \
                        been download), temp files will be stored, and output files will be \
                        written. During StarCluster configuration, all ephemeral drives on the master \
                        node will be combined into a single RAID volume and mounted at the data directory. \
                        Not necessary if '-o', '-t' and '-i' are provided. \
                        Default is '/data'.")
    parser.add_argument('-i', '--input', dest='input', default=None,
                        help="The input file or directory, to be split and processed in parallel. \
                        If a directory is given, all files in the directory will be iteratively processed. \
                        Required, unless <project> is provided.")
    parser.add_argument('-o', '--output', dest='output', default=None,
                        help="The output directory, into which the JSON-formatted output files will be deposited. \
                        If the directory does not exist, it will be created. \
                        Required, unless <project> is provided.")
    parser.add_argument('-l', '--log', dest='log', default=None,
                        help="The log directory, into which log files will be deposited. \
                        Default is <project_directory>/log if <project> is supplied, otherwise \
                        the parent directory of <output> will be used.")
    parser.add_argument('-t', '--temp', dest='temp', default=None,
                        help="The directory in which temp files will be stored. If the directory doesn't exist, \
                        it will be created. Required, unless <project> is provided.")
    parser.add_argument('--sequences', dest='sequences', default=None,
                        help="Only used when passing sequences directly through the API.")
    parser.add_argument('-a', '--assigner', dest='assigner', default='blastn',
                        help='VDJ germline assignment method to use. \
                        Options are: {}. Default is blastn'.format(', '.join(ASSIGNERS.keys())))
    parser.add_argument('-k', '--chunksize', dest='chunksize', default=500, type=int,
                        help="Approximate number of sequences in each distributed job. \
                        Defaults to 500. \
                        Set to 0 if you want file splitting to be turned off \
                        Don't change unless you know what you're doing.")
    parser.add_argument('-O', '--output-type', dest="output_type", action='append',
                        choices=['json', 'imgt', 'minimal'],
                        help="Select the output type. Options are 'json', 'imgt' and 'minimal'. \
                        IMGT output mimics the Summary table produced by IMGT High-V/Quest, \
                        to maintain some level of compatibility with existing IMGT-based pipelines. \
                        JSON output is much more detailed, and is suitable for direct import into MongoDB. \
                        Minimal output is in CSV format. \
                        Defaults to JSON output.")
    parser.add_argument('-m', '--merge', dest="merge", action='store_true', default=False,
                        help="Use if the input files are paired-end FASTQs \
                        (either gzip compressed or uncompressed) from Illumina platforms. \
                        Prior to running the germline assignment pipeline, \
                        paired reads will be merged with PANDAseq.")
    parser.add_argument('-P', '--pandaseq-algo', dest="pandaseq_algo", default='simple_bayesian',
                        choices=['simple_bayesian', 'ea_util', 'flash', 'pear', 'rdp_mle', 'stitch', 'uparse'],
                        help="Define merging algorithm to be used by PANDAseq.\
                        Options are 'simple_bayesian', 'ea_util', 'flash', 'pear', 'rdp_mle', 'stitch', or 'uparse'.\
                        Default is 'simple_bayesian', which is the default PANDAseq algorithm.")
    parser.add_argument('-n', '--nextseq', dest="nextseq", action='store_true', default=False,
                        help="Use if the run was performed on a NextSeq sequencer.")
    parser.add_argument('-u', '--uid', dest="uid", type=int, default=0,
                        help="Length of the unique identifier (UID, or molecular barcode) if used. \
                        Default is unbarcoded (UID length of 0).")
    parser.add_argument('-I', '--isotype', dest="isotype", action='store_false', default=True,
                        help="If set, the isotype will not be determined for heavy chains.\
                        If not set, isotyping sequences for the appropriate species will be used.")
    parser.add_argument('-b', '--basespace', dest="basespace", default=False, action='store_true',
                        help="Use if files should be downloaded directly from BaseSpace. \
                        Files will be downloaded into the input directory.")
    parser.add_argument('-c', '--cluster', dest="cluster", default=False, action='store_true',
                        help="Use if performing computation on a Celery cluster. \
                        If set, input files will be split into many subfiles and passed \
                        to a Celery queue. If not set, input files will still be split, but \
                        will be distributed to local processors using multiprocessing.")
    parser.add_argument('-D', '--debug', dest="debug", action='store_true', default=False,
                        help="If set, logs information about successfully assigned sequences as well \
                        as unsuccessful sequences. Useful for debugging small test datasets, \
                        as the logging is fairly detailed. \
                        Default is to only log unsuccessful sequences.")
    parser.add_argument('-T', '--use-test-data', dest='use_test_data', action='store_true', default=False,
                        help="If set, AbStar will run on a small sample of test data (1000 sequences). \
                        These samples are of human origin, but AbStar doesn't force the use of the human \
                        germline database. Of the 1000 sequences, only 999 contain an antibody rearrangement, \
                        so one sequence should fail the germline assignment process. \
                        Default is False. \
                        Providing temp and output directories (or a project directory) is required.")
    parser.add_argument('-s', '--species', dest='species', default='human',
                        choices=['human', 'macaque', 'mouse', 'rabbit', 'b12mouse', 'vrc01mouse', '9114mouse'])
    parser.add_argument('-z', '--gzip', dest='gzip', default=False, action='store_true',
                        help="Compress the output to a gzipped file")
    parser.add_argument('--raw', dest='raw', default=False, action='store_true',
                        help='Returns raw output (a dict).')
    parser.add_argument('--pretty', dest='pretty', default=False, action='store_true',
                        help='Pretty format json file')
    parser.add_argument('--no-padding', dest='padding', default=True, action='store_false',
                        help="If passed, will eliminate padding from json file. \
                        Don't use if you don't know what you are doing")
    if print_help:
        parser.print_help()
    else:
        args = parser.parse_args()
        return args


class Args(object):
    def __init__(self, project_dir=None, input=None, output=None, log=None, temp=None,
                 sequences=None, chunksize=500, output_type=['json', ], assigner='blastn',
                 merge=False, pandaseq_algo='simple_bayesian', use_test_data=False,
                 nextseq=False, uid=0, isotype=False, pretty=False,
                 basespace=False, cluster=False, padding=True, raw=False,
                 debug=False, species='human', gzip=False):
        super(Args, self).__init__()
        self.sequences = sequences
        self.project_dir = os.path.abspath(project_dir) if project_dir is not None else project_dir
        self.input = os.path.abspath(input) if input is not None else input
        self.output = os.path.abspath(output) if output is not None else output
        self.use_test_data = use_test_data
        self.log = os.path.abspath(log) if log is not None else log
        self.temp = os.path.abspath(temp) if temp is not None else temp
        self.chunksize = int(chunksize)
        self.output_type = [output_type, ] if output_type in [str, unicode] else output_type
        self.assigner = assigner
        self.merge = True if basespace else merge
        self.pandaseq_algo = str(pandaseq_algo)
        self.nextseq = nextseq
        self.uid = int(uid)
        self.isotype = isotype
        self.basespace = basespace
        self.cluster = cluster
        self.pretty = pretty
        self.debug = debug
        self.gzip = gzip
        self.raw = raw
        self.padding = padding
        self.species = species


def validate_args(args):
    if not any([args.project_dir,
                args.sequences,
                all([any([args.input, args.use_test_data]), args.output, args.temp])]):
        parse_arguments(print_help=True)
        sys.exit(1)
    if args.sequences:
        args.output_type = ['json', ]
        args.raw = True
    if all([args.sequences is not None, args.temp is None]):
        args.temp = '/tmp'
    if not args.output_type:
        args.output_type = ['json', ]


#####################################################################
#
#                        FILES AND DIRECTORIES
#
#####################################################################


def make_directories(args):
    full_paths = []
    if args.project_dir is not None and not os.path.exists(args.project_dir):
        _make_direc(args.project_dir, args)
    if args.use_test_data:
        indir = None
    else:
        indir = args.input if args.input is not None else os.path.join(args.project_dir, 'input')
    outdir = args.output if args.output is not None else os.path.join(args.project_dir, 'output')
    tempdir = args.temp if args.temp is not None else os.path.join(args.project_dir, 'temp')
    log_parent = args.project_dir if args.project_dir is not None else os.path.dirname(outdir)
    logdir = args.log if args.log is not None else os.path.join(log_parent, 'log')
    logtempdir = os.path.join(logdir, 'temp')
    for d in [indir, outdir, tempdir, logdir, logtempdir]:
        if d is None:
            full_paths.append(None)
            continue
        d = os.path.abspath(d)
        full_paths.append(d)
        _make_direc(d, args.cluster)
    return full_paths[:-1]


def _make_direc(d, cluster):
    if not os.path.exists(d):
        os.makedirs(d)
    if cluster:
        cmd = 'sudo chmod 777 {}'.format(d)
        p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()


def make_merge_dir(args):
    merge_parent = args.project_dir if args.project_dir is not None else os.path.dirname(indir)
    merge_dir = os.path.abspath(os.path.join(merge_parent, 'merged'))
    _make_direc(merge_dir, args)
    return merge_dir


def setup_logging(log_dir, debug):
    logfile = os.path.join(log_dir, 'abstar.log')
    debug = True if debug > 0 else False
    print_debug = True if debug == 2 else False
    log.setup_logging(logfile, debug=debug)
    global logger
    logger = log.get_logger('abstar')


def log_options(input_dir, output_dir, temp_dir, args):
    logger.info('')
    logger.info('')
    logger.info('-' * 25)
    logger.info('ABSTAR')
    logger.info('-' * 25)
    logger.info('')
    logger.info('SPECIES: {}'.format(args.species))
    logger.info('CHUNKSIZE: {}'.format(args.chunksize))
    logger.info('OUTPUT TYPE: {}'.format(', '.join(args.output_type)))
    if args.merge or args.basespace:
        logger.info('PANDASEQ ALGORITHM: {}'.format(args.pandaseq_algo))
    logger.info('UID: {}'.format(args.uid))
    logger.info('ISOTYPE: {}'.format('yes' if args.isotype else 'no'))
    logger.info('EXECUTION: {}'.format('cluster' if args.cluster else 'local'))
    logger.info('DEBUG: {}'.format('True' if args.debug else 'False'))
    logger.debug('INPUT: {}'.format(input_dir))
    logger.debug('OUTPUT: {}'.format(output_dir))
    logger.debug('TEMP: {}'.format(temp_dir))


def list_files(d, log=False):
    if os.path.isdir(d):
        glob_pattern = os.path.join(os.path.expanduser(d), '*')
        # need to fix square brackets, or glob freaks out
        if any(['[' in glob_pattern, ']' in glob_pattern]):
            glob_pattern = re.sub(r'\[', '[[]', glob_pattern)
            glob_pattern = re.sub(r'(?<!\[)\]', '[]]', glob_pattern)
        files = sorted(glob(glob_pattern))
    else:
        files = [d, ]
    if log:
        fnames = [os.path.basename(f) for f in files]
        logger.info('')
        logger.info('FILE COUNT: {}'.format(len(fnames)))
        logger.info('FILES: {}'.format(', '.join(fnames)))
    return files


def get_output_suffix(output_format):
    osuffixes = {'json': '.json',
                 'imgt': '.csv',
                 'minimal': '.txt'}
    return osuffixes[output_format.lower()]


def build_output_base(output_types):
    needs_header = ['imgt', 'minimal']
    output_dict = {}
    for output_type in output_types:
        if output_type in needs_header:
            output_dict[output_type] = [get_header(output_type), ]
        else:
            output_dict[output_type] = []
    return output_dict


def concat_outputs(input_file, temp_output_files, output_dir, args):
    ofiles = []
    bname = os.path.basename(input_file)
    if '.' in bname:
        split_name = bname.split('.')
        oprefix = '.'.join(split_name[:-1])
    else:
        oprefix = bname
    # temp_output_files is a 2D list in the same order as sorted(args.output_type)
    logger.info('')
    for output_type, temp_files in zip(sorted(args.output_type), zip(*temp_output_files)):
        osuffix = get_output_suffix(output_type)
        oname = oprefix + osuffix
        ofile = os.path.join(output_dir, oname)
        # temp_files = [tf for tf in temp_output_files if tf.endswith(osuffix)]
        logger.info('Concatenating {} {}-formatted job outputs into a single output file'.format(len(temp_output_files),
                                                                                                 output_type.upper()))
        if args.gzip:
            ohandle = gzip.open(ofile + ".gz", 'wb')
        else:
            ohandle = open(ofile, 'w')
        with ohandle as out_file:
            # JSON-formatted files don't have headers, so we don't worry about it
            if output_type == 'json':
                for temp_file in temp_files:
                    with open(temp_file) as f:
                        for line in f:
                            out_file.write(line)
                    out_file.write('\n')
            # For file formats with headers, only keep headers from the first file
            if output_type in ['imgt', 'minimal']:
                for i, temp_file in enumerate(temp_files):
                    with open(temp_file) as f:
                        for j, line in enumerate(f):
                            if i == 0:
                                out_file.write(line)
                            elif j >= 1:
                                out_file.write(line)
                        out_file.write('\n')
        ofiles.append(ofile)
    return ofiles


def concat_logs(input_file, logs, log_dir, log_type):
    bname = os.path.basename(input_file)
    if '.' in bname:
        split_name = bname.split('.')
        lprefix = '.'.join(split_name[:-1])
    else:
        lprefix = bname
    lfile = os.path.join(log_dir, '{}.{}'.format(lprefix, log_type))
    lhandle = open(lfile, 'w')
    with lhandle as logfile:
        for log in logs:
            with open(log) as f:
                for line in f:
                    logfile.write(line)
    return lfile


def clear_temp_files(temp_files):
    for f in temp_files:
        os.unlink(f)


#####################################################################
#
#                       INPUT PROCESSING
#
#####################################################################


def download_files(input_dir):
    from ..utils.basespace import BaseSpace
    bs = BaseSpace()
    logger.info('')
    logger.info('BASESPACE PROJECT NAME: {}'.format(bs.project_name))
    logger.info('BASESPACE PROJECT ID: {}'.format(bs.project_id))
    num_files = bs.download(input_dir)
    logger.info('')
    logger.info('BASESPACE FILE DOWNLOAD COUNT: {}'.format(num_files))


def merge_reads(input_dir, args):
    from ..utils import pandaseq
    merge_dir = make_merge_dir(args)
    pandaseq.run(input_dir,
                 merge_dir,
                 args.pandaseq_algo,
                 args.nextseq)
    return merge_dir


def format_check(input_list):
    formats = []
    for f in input_list:
        formats.append(_get_format(f))
    return formats


def _get_format(in_file):
    with open(in_file) as f:
        line = f.next()
        while line.strip() == '':
            line = f.next()
        if line.lstrip().startswith('>'):
            return 'fasta'
        elif line.lstrip().startswith('@'):
            return 'fastq'
        else:
            return None


def split_file(f, fmt, temp_dir, args):
    file_counter = 0
    seq_counter = 0
    total_seq_counter = 0
    subfiles = []
    sequences = []
    if '.' in os.path.basename(f):
        out_prefix = '.'.join(os.path.basename(f).split('.')[:-1])
    else:
        out_prefix = os.path.basename(f)
    if args.chunksize != 0:
        try:
            with open(f, 'r') as f_handle:
                for seq in SeqIO.parse(f_handle, fmt.lower()):
                    sequences.append(seq.format(fmt.lower()))
                    seq_counter += 1
                    total_seq_counter += 1
                    if seq_counter == args.chunksize:
                        out_file = os.path.join(temp_dir, '{}_{}'.format(out_prefix, file_counter))
                        ohandle = open(out_file, 'w')
                        ohandle.write('\n'.join(sequences))
                        ohandle.close()
                        sequences = []
                        seq_counter = 0
                        file_counter += 1
                        subfiles.append(out_file)
        except ValueError:
            print('')
            print('ERROR: invalid file.')
            print('{} is not properly formatted'.format(f))
            print(traceback.format_exception_only)
            clear_temp_files(subfiles)
            sys.exit(1)

    # We don't want our files split
    else:
        with open(f, 'r') as f_handle:
            for seq in SeqIO.parse(f_handle, fmt.lower()):
                total_seq_counter += 1
            subfiles.append(f)
            file_counter = 1

    # unless the input file is an exact multiple of args.chunksize,
    # need to write the last few sequences to a split file.
    if seq_counter:
        file_counter += 1
        out_file = os.path.join(temp_dir, '{}_{}'.format(out_prefix, file_counter))
        open(out_file, 'w').write('\n' + '\n'.join(sequences))
        subfiles.append(out_file)
    logger.info('SEQUENCES: {}'.format(total_seq_counter))
    logger.info('JOBS: {}'.format(file_counter))
    return subfiles, total_seq_counter


#####################################################################
#
#                            PRINTING
#
#####################################################################


def print_input_file_info(f, fmt):
    fname = os.path.basename(f)
    logger.info('')
    logger.info('')
    logger.info(fname)
    logger.info('-' * len(fname))
    logger.info('FORMAT: {}'.format(fmt.lower()))


def print_job_stats(total_seqs, good_seq_counts, start_time, end_time):
    run_time = end_time - start_time
    zero_files = sum([c == 0 for c in good_seq_counts])
    if zero_files > 0:
        logger.info('{} files contained no successfully processed sequences'.format(zero_files))
    good_seqs = sum(good_seq_counts)
    logger.info('')
    logger.info('{} sequences contained an identifiable rearrangement'.format(good_seqs))
    logger.info('AbStar completed in {} seconds'.format(run_time))



#####################################################################
#
#                             JOBS
#
#####################################################################



@celery.task
def run_abstar(seq_file, output_dir, log_dir, file_format, arg_dict):
    '''
    Wrapper function to multiprocess (or not) the assignment of V, D and J
    germline genes. Also writes the JSON-formatted output to file.

    Input is a a FASTA- or FASTQ-formatted file of antibody sequences, the output directory,
    and a dictionary of runtime args.

    Output is a tuple containing (0) path to the output file, (1) the number of successfully
    annotated antibody sequences, (2) path to the log file for successfully annotated sequences (an
    empty string unless args.debug is True) and (3) path to the log file for unsuccessfully annotated
    sequences (only an eompty string if all sequences in the input file were successful).
    '''
    try:
        # Args instances can't be serialized by Celery, so we need to pass them in
        # as a dict and re-convert to an Args object inside the Celery job
        args = Args(**arg_dict)
        # identify output file
        output_filename = os.path.basename(seq_file)
        output_suffixes = [get_output_suffix(output_type) for output_type in sorted(args.output_type)]
        output_files = [os.path.join(output_dir, output_filename + output_suffix) for output_suffix in output_suffixes]
        # if args.output_type == 'json':
        #     output_file = os.path.join(output_dir, output_filename + '.json')
        # elif args.output_type in ['imgt', 'hadoop']:
        #     output_file = os.path.join(output_dir, output_filename + '.txt')
        # setup log files
        annotated_logfile = os.path.join(log_dir, 'temp/{}.annotated'.format(output_filename))
        annotated_loghandle = open(annotated_logfile, 'a')
        failed_logfile = os.path.join(log_dir, 'temp/{}.failed'.format(output_filename))
        failed_loghandle = open(failed_logfile, 'a')
        unassigned_logfile = os.path.join(log_dir, 'temp/{}.unassigned'.format(output_filename))
        unassigned_loghandle = open(unassigned_logfile, 'a')
        # start assignment
        assigner_class = ASSIGNERS[args.assigner]
        assigner = assigner_class(args.species)  # initialize the assigner class with the species
        assigner(seq_file, file_format)  # call the assigner
        # process all of the successfully assigned sequences
        outputs_dict = build_output_base(args.output_type)
        assigned = [Antibody(vdj, args.species) for vdj in assigner.assigned]
        successful = 0
        for ab in assigned:
            try:
                ab.annotate(args.uid)
                result = get_abstar_result(ab,
                                           pretty=args.pretty,
                                           padding=args.padding,
                                           raw=args.raw)
                for i, output_type in enumerate(args.output_type):
                    output = get_output(result, output_type)
                    if output is not None:
                        outputs_dict[output_type].append(get_output(result, output_type))
                        # only write debug log data once
                        if i == 0:
                            successful += 1
                            if args.debug:
                                annotated_loghandle.write(ab.format_log())
                    # only write failed log data once
                    elif i == 0:
                        failed_loghandle.write(ab.format_log())
            except:
                ab.exception('ANNOTATION ERROR', traceback.format_exc())
                failed_loghandle.write(ab.format_log())
        outputs = [outputs_dict[ot] for ot in sorted(args.output_type)]
        write_output(outputs, output_files)
        # capture the log for all unsuccessful sequences
        for vdj in assigner.unassigned:
            unassigned_loghandle.write(vdj.format_log())
        # close the log handles
        unassigned_loghandle.close()
        annotated_loghandle.close()
        failed_loghandle.close()
        # return the number of successful assignments
        return (output_files, successful, annotated_logfile, failed_logfile, unassigned_logfile)
    except:
        logging.debug(traceback.format_exc())



def process_sequences(sequences, args):
    seq_file = tempfile.NamedTemporaryFile(dir=args.temp, delete=False)
    seq_file.write('\n'.join([s.fasta for s in sequences]))
    seq_file.close()
    assigner_class = ASSIGNERS[args.assigner]
    assigner = assigner_class(args.species)
    assigner(seq_file.name, 'fasta')
    # process all of the successfully assigned sequences
    outputs = []
    assigned = [Antibody(vdj, args.species) for vdj in assigner.assigned]
    for ab in assigned:
        try:
            ab.annotate(args.uid)
            result = get_abstar_result(ab,
                                       pretty=False,
                                       padding=False,
                                       raw=True)
            output = get_output(result, 'json')
            if output is not None:
                outputs.append(get_output(result, 'json'))
        except:
            continue
    os.unlink(seq_file.name)
    return outputs


def run_jobs(files, output_dir, log_dir, file_format, args):
    sys.stdout.write('\nRunning VDJ...\n')
    if args.cluster:
        return _run_jobs_via_celery(files, output_dir, log_dir, file_format, args)
    elif args.debug or args.chunksize == 0:
        return _run_jobs_singlethreaded(files, output_dir, log_dir, file_format, args)
    else:
        return _run_jobs_via_multiprocessing(files, output_dir, log_dir, file_format, args)


def _run_jobs_singlethreaded(files, output_dir, log_dir, file_format, args):
    results = []
    update_progress(0, len(files))
    for i, f in enumerate(files):
        try:
            results.append(run_abstar(f, output_dir, log_dir, file_format, vars(args)))
            update_progress(i + 1, len(files))
        except:
            logger.debug('FILE-LEVEL EXCEPTION: {}'.format(f))
            logging.debug(traceback.format_exc())
    logger.info('')
    return results


def _run_jobs_via_multiprocessing(files, output_dir, log_dir, file_format, args):
    p = Pool(maxtasksperchild=50)
    async_results = []
    update_progress(0, len(files))
    for f in files:
        async_results.append((f, p.apply_async(run_abstar, (f,
                                                         output_dir,
                                                         log_dir,
                                                         file_format,
                                                         vars(args)))))
    monitor_mp_jobs([ar[1] for ar in async_results])
    results = []
    for a in async_results:
        try:
            results.append(a[1].get())
        except:
            logger.debug('FILE-LEVEL EXCEPTION: {}'.format(a[0]))
            # if args.debug:
            #     traceback.print_exc()
            logging.debug(''.join(traceback.format_exc()))
            continue
    p.close()
    p.join()
    return results


def monitor_mp_jobs(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        ready = [ar for ar in results if ar.ready()]
        finished = len(ready)
        update_progress(finished, jobs)
    sys.stdout.write('\n\n')


def _run_jobs_via_celery(files, output_dir, log_dir, file_format, args):
    async_results = []
    for f in files:
        async_results.append(run_abstar.delay(f,
                                           output_dir,
                                           log_dir,
                                           file_format,
                                           vars(args)))
    succeeded, failed = monitor_celery_jobs(async_results)
    failed_files = [f for i, f in enumerate(files) if async_results[i].failed()]
    for ff in failed_files:
        logger.debug('FAILED FILE: {}'.format(f))
    return [s.get() for s in succeeded]


def monitor_celery_jobs(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        succeeded = [ar for ar in results if ar.successful()]
        failed = [ar for ar in results if ar.failed()]
        finished = len(succeeded) + len(failed)
        update_progress(finished, jobs, failed=len(failed))
    sys.stdout.write('\n\n')
    return succeeded, failed


def update_progress(finished, jobs, failed=None):
    pct = int(100. * finished / jobs)
    ticks = pct / 2
    spaces = 50 - ticks
    if failed:
        prog_bar = '\r({}/{}) |{}{}|  {}% ({}, {})'.format(finished, jobs, '|' * ticks, ' ' * spaces, pct, finished - failed, failed)
    else:
        prog_bar = '\r({}/{}) |{}{}|  {}%'.format(finished, jobs, '|' * ticks, ' ' * spaces, pct)
    sys.stdout.write(prog_bar)
    sys.stdout.flush()



#####################################################################
#
#                             MAIN
#
#####################################################################



def run(*args, **kwargs):
    '''
    Runs AbStar.

    Input sequences can be provided in several different formats:

        1) individual sequences as positional arguments: ``run(seq1, seq2, temp=temp, output=output)``
        2) a list of sequences, as an argument: ``run([seq1, seq2], temp=temp, output=output)``
        3) a single FASTA/Q-formatted input file, passed via ``input``
        4) a directory of FASTA/Q-formatted files, passed via ``input``

    When passing sequences (not FASTA/Q files), the sequences can be in any format recognized
    by ``abtools.sequence.Sequence``, including:

        - a raw nucleotide sequence, as a string (a random sequence ID will be assigned)
        - a list/tuple of the format ``[sequence_id, sequence]``
        - a BioPython SeqRecord object
        - an AbTools Sequence object

    Either sequences, ``project_dir``, or all of ``input``, ``output`` and ``temp`` are required.


    Examples:

        If processing a single sequence, you can pass the raw sequence, as a string::

            import abstar

            result = abstar.run('ATGC')

        or a list/tuple of the format ``[sequence_id, sequence]``::

            result = abstar.run(['seq1', 'ATGC'])

        If you pass just the raw sequence, a random sequence ID will be generated with ``uuid.uuid4()``.
        In either case, when given a single sequence, ``abstar.run()`` will return a single AbTools ``Sequence``
        object. If running multiple sequences, you can either pass each sequence as a positional argument::

            result_list = run(['seq1', 'ATGC'], ['seq2', 'CGTA'])

        or you can pass a list of sequences as the first argument, in this case using sequences parsed from a
        FASTA file using Biopython::

            from Bio import SeqIO

            fasta = open('my_sequences.fasta', 'r')
            seqs = [s for s in SeqIO.parse(fasta, 'fasta')]
            result_list = abstar.run(seqs)

        When given multiple sequences, ``abstar.run()`` will return a list of AbTools ``Sequence`` objects,
        one per input sequence.

        If you'd prefer not to parse the FASTQ/A file into a list (for example, if the input file is
        extremely large), you can pass the input file path directly, along with a temp directory and output
        directory::

            result_files = abstar.run(input='/path/to/my_sequences.fasta',
                                      temp='/path/to/temp',
                                      output='/path/to/output')

        Given a file path, ``abstar.run()`` returns a list of output file paths. In the above case,
        ``result_files`` will be a list containing a single output file path:
        ``/path/to/output/my_sequences.json``.

        If you have a directory containing multiple FASTQ/A files, you can pass the directory path
        using ``input``::

            result_files = abstar.run(input='/path/to/input',
                                      temp='/path/to/temp',
                                      output='/path/to/output')

        As before, ``result_files`` will contain a list of output file paths.

        If your input directory contains paired FASTQ files (gzip compressed or uncompressed)
        that need to be merged prior to processing with AbStar::

            result_files = abstar.run(input='/path/to/input',
                                      temp='/path/to/temp',
                                      output='/path/to/output',
                                      merge=True)

        The paired read files in ``input`` will be merged with PANDAseq prior to processing with AbStar.
        By default, PANDAseq's 'simple bayesian' read merging algorithm is used, although alternate
        algorithms can be selected with ``pandaseq_algo``.

        AbStar also provides an alternate CSV-formatted output type that mimics the `IMGT Summary file`_.
        This option is provided to minimize the effort needed to convert existing
        IMGT-based pipelines to AbStar. Alternate output is only available when passing an input file or
        directory; passing individual sequences or a list of sequences will always return Sequence objects.
        To produce IMGT-formatted output::

            result_files = abstar.run(input='/path/to/input',
                                      temp='/path/to/temp',
                                      output='/path/to/output',
                                      output_type='imgt')

        .. _IMGT Summary file: http://www.imgt.org/IMGT_vquest/share/textes/imgtvquest.html#Esummary


    Args:

        project_dir (str): Path to the project directory. Most useful when directly downloading
            files from BaseSpace, and all subdirectories will be created by AbStar.

        input (str): Path to input directory, containing FASTA/Q files. If performing
            read merging with PANDAseq, paired FASTQ files may be gzip compressed.

        output (str): Path to output directory.

        temp (str): Path to temp directory, where intermediate job files will be stored.

        log (str): Path to log file. If not provided and ``project_dir`` is provided,
            the log will be written to ``/path/to/project_dir/abstar.log``. If output is
            provided, log will be written to ``/path/to/output/abstar.log``.

        species (str): Species of the antibody sequences. Choices are 'human', 'macaque',
            'mouse' and 'rabbit'. Default is 'human'.

        isotype (bool): If True, the isotype will infered by aligning the sequence region
            downstream of the J-gene. If False, the isotype will not be determined.
            Default is True.

        uid (int): Length (in nucleotides) of the Unique Molecular ID used to barcode input RNA.
            A positive integer results in the UMID being parsed from the start of the read (or merged
            read), a negative integer results in parsing from the end of the read. Default is 0,
            which results in no UMID parsing.

        gzip (bool): If True, compresses output files with gzip. Default is False.

        pretty (bool): If True, formats JSON output files to be more human-readable. If False,
            JSON output files contain one record per line. Default is False.

        output_type (str): Options are 'json' or 'imgt'. IMGT output mimics the Summary
            table produced by IMGT High-V/Quest, to maintain a level of compatibility with
            existing IMGT-based pipelines. JSON output is much more detailed. Default is 'json'.

        merge (bool): If True, input must be paired-read FASTA files (gzip compressed or uncompressed)
            which will be merged with PANDAseq prior to processing with AbStar. If ``basespace`` is True,
            ``merge`` is automatically set to True. Default is False.

        pandaseq_algo (str): Define merging algorithm to be used by PANDAseq. Options are
            'simple_bayesian', 'ea_util', 'flash', 'pear', 'rdp_mle', 'stitch', or 'uparse'. Default is
            'simple_bayesian', which is the default PANDAseq algorithm.

        debug (bool): If ``True``, ``abstar.run()`` runs in single-threaded mode, the log is much more verbose,
            and temporary files are not removed. Default is ``False``.


    Returns:

        If the input is a single sequence, ``run`` returns a single AbTools ``Sequence`` object.

        If the input is a list of sequences, ``run`` returns a list of AbTools ``Sequence`` objects.

        If the input is a file or a directory of files, ``run`` returns a list of output files.
    '''

    warnings.filterwarnings("ignore")
    if len(args) == 1:
        # if there's a single arg, need to check if it's a single sequence...
        try:
            sequences = [Sequence(args[0]), ]
        except:
            # ...or a list of sequences
            try:
                sequences = [Sequence(s) for s in args[0]]
            except:
                print('ERROR: invalid format for sequence input:')
                for a in args:
                    print(a)
                sys.exit(1)
    # if multiple args, assume each is a sequence
    elif len(args) > 1:
        try:
            sequences = [Sequence(s) for s in args]
        except:
            print('ERROR: invalid format for sequence input:')
            for a in args:
                print(a)
            sys.exit(1)
    kwargs['sequences'] = sequences
    args = Args(**kwargs)
    validate_args(args)
    global logger
    logger = log.get_logger('abstar')
    output = main(args)
    # if args.sequences is not None:
    #     output = [Sequence(o) for o in output]
    #     if len(output) == 1:
    #         return output[0]
    return output


def run_standalone(args):
    output_dir = main(args)


def main(args):
    if args.sequences is not None:
        # from utils import output, vdj
        processed = process_sequences(args.sequences, args)
        if len(processed) == 1:
            return Sequence(dict(processed[0]))
        return [Sequence(dict(p)) for p in processed]
    else:
        input_dir, output_dir, temp_dir, log_dir = make_directories(args)
        setup_logging(log_dir, args.debug)
        log_options(input_dir, output_dir, temp_dir, args)
        if args.use_test_data:
            mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            input_files = [os.path.join(mod_dir, 'test_data/test_1k.fasta'), ]
        else:
            if args.basespace:
                args.merge = True
                download_files(input_dir)
            if args.merge:
                input_dir = merge_reads(input_dir, args)
            if args.isotype:
                args.isotype = args.species
            input_files = [f for f in list_files(input_dir, log=True) if os.stat(f).st_size > 0]
        output_files = []
        assigned_files = []
        unassigned_files = []
        for f, fmt in zip(input_files, format_check(input_files)):
            # skip the non-FASTA/Q files
            if fmt is None:
                continue
            start_time = time.time()
            print_input_file_info(f, fmt)
            subfiles, seq_count = split_file(f, fmt, temp_dir, args)
            run_info = run_jobs(subfiles, temp_dir, log_dir, fmt, args)
            temp_output_files = [r[0] for r in run_info if r is not None]
            processed_seq_counts = [r[1] for r in run_info if r is not None]
            annotated_log_files = [r[2] for r in run_info if r is not None]
            failed_log_files = [r[3] for r in run_info if r is not None]
            unassigned_log_files = [r[4] for r in run_info if r is not None]
            vdj_end_time = time.time()
            _output_files = concat_outputs(f, temp_output_files, output_dir, args)
            unassigned_file = concat_logs(f, unassigned_log_files, log_dir, 'unassigned')
            failed_file = concat_logs(f, failed_log_files, log_dir, 'failed')
            if args.debug:
                annotated_file = concat_logs(f, annotated_log_files, log_dir, 'annotated')
            output_files.extend(output_files)
            if not args.debug:
                flat_temp_files = [f for subl in temp_output_files for f in subl]
                clear_temp_files(subfiles + flat_temp_files + annotated_log_files + failed_log_files + unassigned_log_files)
            print_job_stats(seq_count, processed_seq_counts, start_time, vdj_end_time)
        return output_files


if __name__ == '__main__':
    warnings.filterwarnings("ignore")
    args = parse_arguments()
    output_dir = main(args)
    sys.stdout.write('\n\n')



#!/usr/bin/python
# filename: productivity.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import traceback


def check_productivity(antibody):
    return Productivity(antibody)


class Productivity(object):
    """docstring for Productivity"""
    def __init__(self, antibody):
        super(Productivity, self).__init__()
        self.antibody = antibody
        self.productivity_issues = []
        self._is_productive = None


    @property
    def is_productive(self):
        if self._is_productive is None:
            try:
                problems = any([self.stop_codons(self.antibody),
                                self.ambig_codons(self.antibody),
                                self.vdj_disagreement(self.antibody),
                                self.missing_conserved_junc_residues(self.antibody),
                                self.junction_not_in_frame(self.antibody),
                                self.out_of_frame_indels(self.antibody),
                                ])
                if problems:
                    self._is_productive = False
                else:
                    self._is_productive = True
            except:
                self.antibody.exception('PRODUCTIVITY ERROR', traceback.format_exc())
                raise
        return self._is_productive


    def stop_codons(self, antibody):
        if '*' in antibody.vdj_aa:
            self.productivity_issues.append('Contains stop codon(s)')
            return True
        return False


    def ambig_codons(self, antibody):
        if 'X' in antibody.vdj_aa.upper():
            self.productivity_issues.append('Contains ambiguous codon(s)')
            return True
        return False


    def vdj_disagreement(self, antibody):
        if antibody.v.full[:3] != antibody.j.full[:3]:
            self.productivity_issues.append('V-gene ({}) and J-gene({}) disagree'.format(antibody.v.gene, antibody.j.gene))
            return True
        if antibody.d:
            if not antibody.d.gene:
                return False
            if antibody.v.full[:3] != antibody.d.full[:3]:
                self.productivity_issues.append('V-gene ({}) and D-gene({}) disagree'.format(antibody.v.gene, antibody.d.gene))
                return True
        return False


    def missing_conserved_junc_residues(self, antibody):
        if not antibody.junction:
            return True
        start = 'C'
        end = 'W' if antibody.chain == 'heavy' else 'F'
        if antibody.junction.junction_aa[0] != start or antibody.junction.junction_aa[-1] != end:
            self.productivity_issues.append('Junction ({}) lacks conserved start and/or end residue'.format(antibody.junction.junction_aa))
            return True
        return False


    def junction_not_in_frame(self, antibody):
        if not antibody.junction.in_frame:
            self.productivity_issues.append('Junction ({}) is out of frame'.format(antibody.junction.junction_nt))
            return True
        return False


    def out_of_frame_indels(self, antibody):
        problems = 0
        for indels in [antibody.v.insertions, antibody.j.insertions, antibody.v.deletions, antibody.j.deletions]:
            for i in indels:
                if i.in_frame == 'no':
                    self.productivity_issues.append('Contains a frameshift {}: {}'.format(i.type, i.abstar_formatted))
                    problems += 1
        return False if problems == 0 else True



#!/usr/bin/python
# filename: make_basespace_credfile.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function, absolute_import

import json
import platform
import os
import sys


def main():
    creds = {}
    print('\nEnter your BaseSpace credentials')
    creds['client_id'] = raw_input('Client ID: ')
    if not creds['client_id']:
        print('ERROR: Client ID is a required field.')
        sys.exit(1)
    creds['client_secret'] = raw_input('Client Secret: ')
    if not creds['client_secret']:
        print('ERROR: Client Secret is a required field.')
        sys.exit(1)
    creds['access_token'] = raw_input('Access Token: ')
    if not creds['access_token']:
        print('ERROR: Access Token is a required field.')
        sys.exit(1)
    print("\nThe following options should only be provided if you know what you're doing.")
    print("If unsure, leave blank to use the default value.")
    creds['api_server'] = raw_input('API Server [https://api.basespace.illumina.com/]: ')
    if not creds['api_server'].strip():
        creds['api_server'] = 'https://api.basespace.illumina.com/'
    creds['version'] = raw_input('Version [v1pre3]: ')
    if not creds['version'].strip():
        creds['version'] = 'v1pre3'
    write_cred_file(creds)


def write_cred_file(creds):
    jcreds = json.dumps(creds)
    cred_file = os.path.expanduser('~/.abstar/basespace_credentials')
    if not os.path.exists(os.path.dirname(cred_file)):
        os.makedirs(os.path.dirname(cred_file))
    cred_handle = open(cred_file, 'w')
    cred_handle.write(jcreds)
    print('\nCredentials have been saved to {}\n'.format(cred_file))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# filename: partis.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import csv
import os
import platform
import subprocess as sp
from tempfile import NamedTemporaryFile
import traceback

from abtools.alignment import local_alignment
from abtools.sequence import Sequence

from .assigner import BaseAssigner
from ..core.germline import GermlineSegment
from ..core.vdj import VDJ



class Partis(BaseAssigner):
    """
    docstring for Partis
    """

    def __init__(self, species):
        super(Partis, self).__init__(species)


    def __call__(self, sequence_file, file_format, locus='ig'):
        self.assigned, self.unassigned = run_partis(sequence_file, file_format, locus)


    def run_partis(self, sequence_file, file_format, locus):
        with open(sequence_file, 'r') as sequence_handle:
            seqs = [Sequence(s) for s in SeqIO.parse(sequence_handle, file_format)]
        seq_dict = {s.id.replace(':', 'c'): s for s in seqs}
        partis_out = NamedTemporaryFile(delete=False)
        germline_dir = os.path.join(self.germline_directory, 'partis/')
        locus = 'igh' if locus == 'ig' else 'tra'
        partis_cmd = ['partis', 'run-viterbi',
                      '--infname', sequence_file,
                      '--outfname', partis_out.name,
                      '--locus', locus,
                      '--initial-germline-dir', germline_dir]
        p = sp.Popen(partis_cmd, stdout=sp.PIPE, stderr=sp.PIPE)
        stdout, stderr = p.communicate()
        parsed_vdjs = self.parse_partis_output(partis_out.name, seq_dict)
        os.unlink(partis_out.name)
        return parsed_vdjs


    def parse_partis_output(self, output_file, seq_dict):
        assigned = []
        unassigned = []
        with open(output_file, 'rb') as f:
            csv_dict = list(csv.DictReader(f))
        for c in csv_dict:
            partis_id = c['unique_ids']
            seq = seq_dict[partis_id]
            seq_id = seq.id
            v = c['v_gene']
            if v[:3].upper() in ['IGH', 'TRA', 'TRD']:
                d = c['d_gene'] if c['d_gene'].strip() else None
            else:
                d = None
            j = c['j_gene']
            vdjs.append(VDJ(seq, v=v, d=d, j=j))
            if c['duplicates'].strip():
                for dup_id in c['duplicates'].strip().split(':'):
                    seq = seq_dict[dup_id]
                    vdjs.append(VDJ(seq, v=v, d=d, j=j))
        return vdjs


    def compute_alignment_positions(self):
        pass



#!/usr/bin/env python
# filename: antibody.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import traceback

from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from abtools.alignment import global_alignment, local_alignment

from ..utils import isotype, junction, mutations, productivity, regions
from ..utils.mixins import LoggingMixin
from .germline import get_imgt_germlines


class Antibody(object, LoggingMixin):
    """
    docstring for Antibody
    """
    def __init__(self, vdj, species):
        super(Antibody, self).__init__()
        LoggingMixin.__init__(self)
        self.id = vdj.sequence.id
        self.raw_input = vdj.sequence
        self.oriented_input = vdj.oriented
        self.v = vdj.v
        self.j = vdj.j
        self.d = vdj.d
        self.chain = vdj.v.chain
        self.species = species.lower()
        # initialize the log
        self.initialize_log()
        # property vars
        self._strand = None
        self._imgt_position_from_aligned = {}
        self._aligned_position_from_imgt = {}


    @property
    def strand(self):
        if self._strand is None:
            if self.v is None:
                self._strand = None
            else:
                self._strand = self.v.strand
        return self._strand


    @property
    def has_exceptions(self):
        if len(self.exceptions) > 0:
            return True
        segs = [s for s in [self.v, self.d, self.j] if s is not None]
        if any([len(seg.exceptions) > 0 for seg in segs]):
            return True
        return False


    def initialize_log(self):
        log = []
        log.append('=' * len(self.id))
        log.append(self.id)
        log.append('=' * len(self.id))
        log.append('')
        log.append(self.raw_input.fasta)
        log.append('')
        log.append('RAW INPUT: {}'.format(self.raw_input.sequence))
        log.append('ORIENTED INPUT: {}'.format(self.oriented_input.sequence))
        log.append('CHAIN: {}'.format(self.chain))
        log.append('')
        if self.v is not None:
            log.append('V-GENE: {}'.format(self.v.full))
        if self.d is not None:
            log.append('D-GENE: {}'.format(self.d.full))
        if self.j is not None:
            log.append('J-GENE: {}'.format(self.j.full))
        self._log = log



    def annotate(self, uid):
        '''
        Realigns the V(D)J germline genes using optimized SSW alignment
        parameters, followed by more detailed annotation (junction
        identification, ambig correction, parsing of mutations and
        indels, etc).
        '''
        try:
            # print(self.id)
            # print('Parsing UIDs...')
            self._parse_uid(uid)
            # print('Realigning germlines...')
            self._realign_germlines()
            # print('Processing junction...')
            self._get_junction()
            # print('Assembling VDJ sequence...')
            self._assemble_vdj_sequence()
            # print('Identifying regions...')
            self._identify_regions()
            # print('Mutations...')
            self._mutations()
            # print('Isotypes...')
            self._isotype()
            # print('Productivity...')
            self._productivity()
        except:


            # print(traceback.format_exc())

            self.exception('ANNOTATION', traceback.format_exc())


# ---------------------
# GERMLINE REALIGNMENT
# ---------------------

    def _realign_germlines(self):
        # realign V and compute gapped IMGT alignment for the V gene
        self.log('')
        self.log('V-GENE REALIGNMENT')
        self.log('------------------')
        self.v.realign_germline(self)
        self.log('RAW QUERY SEQUENCE:', self.v.raw_query)
        self.log('RAW QUERY LENGTH:', len(self.v.raw_query))
        self.log('RAW GERMLNE LENGTH:', len(self.v.raw_germline))
        self.log('QUERY START:', self.v.query_start)
        self.log('QUERY END:', self.v.query_end)
        self.log('GERMLINE START:', self.v.germline_start)
        self.log('GERMLINE END:', self.v.germline_end)
        self.log('')
        self.log('  QUERY: ', self.v.query_alignment)
        self.log('         ', self.v.alignment_midline)
        self.log('GERMLINE:', self.v.germline_alignment)
        self.log('')
        self.log('V-GENE GAPPED IMGT ALIGNMENT')
        self.log('----------------------------')
        self.v.gapped_imgt_realignment()
        self.log('  QUERY: ', self.v.imgt_gapped_alignment.aligned_query)
        self.log('         ', self.v.imgt_gapped_alignment.alignment_midline)
        self.log('GERMLINE:', self.v.imgt_gapped_alignment.aligned_target)
        self.log('')
        self.log('IMGT ALIGNMENT START:', self.v.imgt_gapped_alignment.target_begin)
        self.log('IMGT GERMLINE CODING START:', self.v.imgt_germline.coding_start)
        self.log('V-GENE ALIGNMENT READING FRAME:', self.v.alignment_reading_frame)
        self.log('V-GENE CODING REGION:', self.v.coding_region)
        self.log('V-GENE AA SEQUENCE:', self.v.aa_sequence)

        # realign J
        jstart = self.v.query_end + 1
        self.log('')
        self.log('J-GENE REALIGNMENT')
        self.log('------------------')
        self.j.realign_germline(self, query_start=jstart)
        self.log('RAW QUERY SEQUENCE:', self.j.raw_query)
        self.log('RAW QUERY LENGTH:', len(self.j.raw_query))
        self.log('RAW GERMLNE LENGTH:', len(self.j.raw_germline))
        self.log('QUERY START:', self.j.query_start)
        self.log('QUERY END:', self.j.query_end)
        self.log('GERMLINE START:', self.j.germline_start)
        self.log('GERMLINE END:', self.j.germline_end)
        self.log('')
        self.log('  QUERY: ', self.j.query_alignment)
        self.log('         ', self.j.alignment_midline)
        self.log('GERMLINE:', self.j.germline_alignment)
        self.log('')
        self.log('J-GENE GAPPED IMGT ALIGNMENT')
        self.log('----------------------------')
        self.j.gapped_imgt_realignment()
        self.log('  QUERY: ', self.j.imgt_gapped_alignment.aligned_query)
        self.log('         ', self.j.imgt_gapped_alignment.alignment_midline)
        self.log('GERMLINE:', self.j.imgt_gapped_alignment.aligned_target)
        self.log('')
        self.log('IMGT ALIGNMENT START:', self.j.imgt_gapped_alignment.target_begin)
        self.log('IMGT GERMLINE CODING START:', self.j.imgt_germline.coding_start)
        self.log('J-GENE ALIGNMENT READING FRAME:', self.j.alignment_reading_frame)
        self.log('J-GENE CODING REGION:', self.j.coding_region)
        self.log('J-GENE AA SEQUENCE:', self.j.aa_sequence)

        # realign D (if needed)
        if all([self.chain == 'heavy', self.d is not None]):
            try:
                dstart = self.v.query_end + 1
                dend = self.j.query_start
                self.log('')
                self.log('D-GENE REALIGNMENT')
                self.log('------------------')
                self.d.realign_germline(self, query_start=dstart, query_end=dend)
                self.log('RAW QUERY LENGTH:', len(self.d.raw_query))
                self.log('RAW GERMLNE LENGTH:', len(self.d.raw_germline))
                self.log('QUERY START:', self.d.query_start)
                self.log('QUERY END:', self.d.query_end)
                self.log('GERMLINE START:', self.d.germline_start)
                self.log('GERMLINE END:', self.d.germline_end)
                self.log('  QUERY: ', self.d.query_alignment)
                self.log('         ', self.d.alignment_midline)
                self.log('GERMLINE:', self.d.germline_alignment)
            except:
                self.exception('D-GENE REALIGNMENT ERROR', traceback.format_exc())
                self.d = None



# -------------
# VDJ ASSEMBLY
# -------------

    def _assemble_vdj_sequence(self):
        self.log('')
        self.log('VDJ ASSEMBLY')
        self.log('------------')
        self.gapped_vdj_nt = self._gapped_vdj_nt()
        self.log('GAPPED VDJ NT SEQUENCE:', self.gapped_vdj_nt)
        self.vdj_nt = self.gapped_vdj_nt.replace('-', '')
        self.log('VDJ NT SEQUENCE:', self.vdj_nt)
        self.vdj_aa = self._vdj_aa()
        self.log('VDJ AA SEQUENCE:', self.vdj_aa)
        self.gapped_vdj_germ_nt = self._gapped_vdj_germ_nt()
        self.log('GAPPED GERMLINE VDJ NT SEQUENCE:', self.gapped_vdj_germ_nt)
        self.vdj_germ_nt = self.gapped_vdj_germ_nt.replace('-', '')
        self.log('GERMLINE VDJ NT SEQUENCE:', self.vdj_germ_nt)
        self.vdj_germ_aa = self._vdj_germ_aa()
        self.log('GERMLINE VDJ AA SEQUENCE:', self.vdj_germ_aa)


    def _gapped_vdj_nt(self):
        'Returns the gapped nucleotide sequence of the VDJ region.'
        try:
            gapped_vdj_nt = self.v.query_alignment + \
                self.oriented_input[self.v.query_end + 1:self.j.query_start] + \
                self.j.query_alignment
            return gapped_vdj_nt
        except:
            self.exception('GAPPED VDJ NT ASSEMBLY', traceback.format_exc())


    def _vdj_aa(self):
        'Returns the amino acid sequence of the VDJ region.'
        self.v_rf_offset = (len(self.oriented_input[self.v.query_start:self.junction.junction_nt_start]) % 3)
        self.coding_start = self.v.query_start + self.v_rf_offset
        self.coding_end = self.j.query_end - (len(self.oriented_input[self.coding_start:self.j.query_end])) % 3
        self.coding_region = self.oriented_input[self.coding_start:self.coding_end + 1]
        translated_seq = Seq(self.coding_region, generic_dna).translate()
        self.log('READING FRAME OFFSET:', self.v_rf_offset)
        self.log('CODING START:', self.coding_start)
        self.log('CODING END:', self.coding_end)
        self.log('CODING REGION:', self.coding_region)
        return str(translated_seq)


    def _gapped_vdj_germ_nt(self):
        'Returns the gapped germline nucleotide sequence of the VDJ region.'
        try:
            if self.d:
                germ_junction = self.junction.n1_nt + self.d.germline_alignment + self.junction.n2_nt
            else:
                germ_junction = self.junction.n_nt
            germ_vdj_nt = self.v.germline_alignment + germ_junction + self.j.germline_alignment
            return germ_vdj_nt
        except:
            self.exception('GAPPED GERMLINE VDJ NT ASSEMBLY', traceback.format_exc())


    def _vdj_germ_aa(self):
        'Returns the germline amino acid sequence of the VDJ region.'
        trim = len(self.vdj_germ_nt) - (len(self.vdj_germ_nt[self.v_rf_offset:]) % 3)
        translated_seq = Seq(self.vdj_germ_nt[self.v_rf_offset:trim], generic_dna).translate()
        return str(translated_seq)


    def _parse_uid(self, uid):
        if uid >= 0:
            self.uid = self.raw_input[:uid]
        else:
            self.uid = self.raw_input[uid:]


    def _get_junction(self):
        self.junction = junction.get_junction(self)


    def _identify_regions(self):
        '''
        Identifies and annotates variable/joining gene regions.
        '''
        # from abstar.utils import regions
        self.v.regions = regions.get_variable_regions(self)
        self.j.regions = regions.get_joining_regions(self)
        self.log('')
        self.log('VDJ REGIONS')
        self.log('-----------')
        self.log('FR1 NT sequence:', self.v.regions.nt_seqs['FR1'])
        self.log('CDR1 NT sequence:', self.v.regions.nt_seqs['CDR1'])
        self.log('FR2 NT sequence:', self.v.regions.nt_seqs['FR2'])
        self.log('CDR2 NT sequence:', self.v.regions.nt_seqs['CDR2'])
        self.log('CDR3 NT sequence:', self.junction.cdr3_nt)
        self.log('FR3 NT sequence:', self.v.regions.nt_seqs['FR3'])
        self.log('FR4 NT sequence:', self.j.regions.nt_seqs['FR4'])
        self.log('REGION POSITIONS:', self.v.regions.raw_nt_positions)
        self.log('FR1 AA sequence:', self.v.regions.aa_seqs['FR1'])
        self.log('CDR1 AA sequence:', self.v.regions.aa_seqs['CDR1'])
        self.log('FR2 AA sequence:', self.v.regions.aa_seqs['FR2'])
        self.log('CDR2 AA sequence:', self.v.regions.aa_seqs['CDR2'])
        self.log('FR3 AA sequence:', self.v.regions.aa_seqs['FR3'])
        self.log('CDR3 AA sequence:', self.junction.cdr3_aa)
        self.log('FR4 AA sequence:', self.j.regions.aa_seqs['FR4'])
        self.log('')


    def _mutations(self):
        '''
        Identifies and annotates nucleotide mutations.
        '''
        self.nt_mutations = mutations.nt_mutations(self)
        self.aa_mutations = mutations.aa_mutations(self)


    def _isotype(self):
        self.isotype = isotype.get_isotype(self)


    def _productivity(self):
        self.productivity = productivity.check_productivity(self)


    @staticmethod
    def _realignment_scoring_params(gene):
        '''
        Returns realignment scoring paramaters for a given gene type.

        Args:

            gene (str): the gene type ('V', 'D', or 'J')


        Returns:

            dict: realignment scoring parameters
        '''
        scores = {'V': {'match': 3,
                        'mismatch': -2,
                        'gap_open': -22,
                        'gap_extend': -1},
                  'D': {'match': 3,
                        'mismatch': -2,
                        'gap_open': -22,
                        'gap_extend': -1},
                  'J': {'match': 3,
                        'mismatch': -2,
                        'gap_open': -22,
                        'gap_extend': -1}}
        return scores[gene]



#!/usr/bin/env python
# filename: mixins.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function


class LoggingMixin():
    """docstring for LoggingMixin"""
    def __init__(self):
        self._log = None
        self._exceptions = None


    @property
    def logs(self):
        if self._log is not None:
            return self._log
        return []


    @property
    def exceptions(self):
        if self._exceptions is not None:
            return self._exceptions
        return []


    def log(self, *args, **kwargs):
        '''
        Records a log message
        '''
        sep = kwargs.get('sep', ' ')
        lstring = sep.join([str(a) for a in args])
        if self._log is None:
            self._log = [lstring, ]
        else:
            self._log.append(lstring)


    def exception(self, *args, **kwargs):
        '''
        Records an exception.
        '''
        sep = kwargs.get('sep', '\n')
        estring = sep.join([str(a) for a in args])
        if self._exceptions is None:
            self._exceptions = [estring, ]
        else:
            self._exceptions.append(estring)


    def format_log(self):
        '''
        Formats the log, including exceptions.

        Log formatting will only be performed on sequences that had an
        error during annotation, unless AbStar is run in debug
        mode. In debug mode, all sequences will be logged.

        Returns:
        --------

            str: Formatted log string.
        '''
        output = ''
        output += '\n'.join(self.logs)
        if self._check_for_exceptions():
            output += '\n\n'
            output += self._format_exceptions()
        output += '\n\n'
        return output


    def _check_for_exceptions(self):
        if self.exceptions:
            return True
        if self.v is not None:
            if self.v._exceptions:
                return True
        if self.d is not None:
            if self.d._exceptions:
                return True
        if self.j is not None:
            if self.j._exceptions:
                return True
        return False


    def _format_exceptions(self):
        exceptions = []
        exceptions += self.exceptions
        estring = '\n\nEXCEPTIONS\n'
        estring += '----------\n\n'
        if self.v is not None:
            exceptions += self.v.exceptions
        if self.d is not None:
            exceptions += self.d.exceptions
        if self.j is not None:
            exceptions += self.j.exceptions
        estring += '\n\n'.join([e for e in exceptions])
        return estring



#!/usr/bin/env python
# filename: vdj.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import abc
import os

from ..core.germline import get_germline_database_directory


class BaseAssigner(object):
    """
    ``BaseAssigner`` provides an abstract base class for custom VDJ Assigners.

    Overview
    --------

    Several different tools exist for inferring the germline components of recombined
    antibody sequences. Due to their differing design goals, these tools often have
    different priorities -- some are optimized for speed with a small (often
    negligible) accuracy penalty, while others define recombination junctions with high
    precision at the cost of speed. Because there isn't a single germline assignment tool
    that is optimal for every use case, the germline assignment component of AbStar has been
    designed to be modular and replaceable, allowing the relatively straightforward addition
    of new germline assignment tools and the selection of the desired Assigner at runtime.
    AbStar's default Assigner (`blastn`) identifies V- and J-genes using BLASTn and identifies
    D-genes using rapid Smith-Waterman local alignment.

    The basic purpose of any custom Assigner class is to accept sequences and
    produce ``VDJ`` objects. ``VDJ`` objects package the input sequence together with
    ``GermlineSegment`` objects that contain information about V, D and/or J assignments.
    Following germline assignment by the Assigner, addtional annotation will be performed
    by AbStar. The ``VDJ`` objects provide a common interface by which various Assigners
    can communicate with the additional sequence annotation components in AbStar. This additional
    annotation is consistent regardless of the Assigner (as is the output schema), and provides
    a unified method by which different germline Assigners can be used across different projects
    while maintaining schematic compatibility with downstream analysis tools.

    ``BaseAssigner`` provides the following attributes:

      - ``assigned``: a list designed to contain ``VDJ`` objects for which
                    successful V(D)J assignment has been completed. These
                    ``VDJ`` objects will be additionally annotated by AbStar
                    following assignment. Exception/log information in these
                    ``VDJ`` objects will only be written to the log file if AbStar
                    is run in debug mode.

      - ``unassigned``: a list designed to contain ``VDJ`` objects for which
                      V(D)J assigjment was unsuccessful. These ``VDJ`` objects
                      will not be additionally annotated by AbStar and any exception/log
                      information contained in these ``VDJ`` objects will be logged
                      (even if AbStar was not run in debug mode).

      - ``germline_directory``: path to the directory containing germline databases.
                              Germline DBs are heirarchically grouped into separate folders
                              first by species, then by the Assigner for which they were
                              designed (``blastn``, for example). Within each folder, germline
                              DB files are named as ``{segment}.fasta``, so the Blast V-gene
                              DB for human would be ``human/blast/v.fasta``. Note that this will
                              change as AbStar is updated to annotate TCRs in addition to antibodies

      - ``binary_directory``: path to the directory containing Assigner binaries. Binaries
                            are named as ``{binary}_{system}``, where ``system`` is the
                            lowercase output from platform.system(). For example, the
                            Blastn binary for OSX would be at ``binary_directory/blastn_darwin``.
                            Some assigners require more than a simple binary for execution (Partis,
                            for example). In this case, all required elements will be placed into
                            a folder within the binary directory, named using the same convention
                            as binaries. So the folder containing Linux-compiled Partis components
                            will be ``binary_directory/partis_linux``. The main Partis executable
                            would then be located at ``binary_directory/partis_linux/bin/partis``.

    In order to build a custom assignment class, you simply need to subclass BaseAssigner, implement
    your assigner, and register your assigner in abstar.assigners.registry.ASSIGNERS. Obviously, this
    is in addition to adding any required binaries and specialized germline database formats. AbStar
    prefers to incorporate assigner binaries if at all possible (as opposed to relying on system
    installations) to smooth the process of installing AbStar and eliminate the need for users to
    install several additional programs before having a fully operational AbStar installation.

    There are two critical elements to the design of your assigner class. First, ensure that
    you call the BaseAssigner's ``__init__()`` at the start of your class's ``__init__()``. Second,
    you must implement the  ``__call__()`` method.


    ``__init__()``
    --------------

    Your custom assigner class's ``__init__()`` must accept a single argument (``species``, in
    addition to the required ``self``). You must then call the BaseAssigner's ``__init__()`` and pass
    the ``species`` argument. The simplest was to do this is by using ``super()``, like this:
    ``super(MyAssigner, self).__init__(species)``, where ``MyAssigner`` is your assigner class.


    Implementing the ``__call__()`` method
    --------------------------------------

      - ``__call__()`` must accept two arguments. The first (``sequence_file``) is a FASTA- or
        FASTQ-formatted file of input sequences, depending on the user-provided input. The second argument
        (``file_format``) defines the format of ``sequence_file``, either ``'fasta'`` or ``'fastq'``. If
        a FASTQ file is required by your assigner, ensure that ``__call__()`` raises the appropriate exception
        (and provides a sufficiently clear description of the problem) if a FASTA-formatted file is provided.

      - The result of ``__call__()`` should be the generation of a ``VDJ`` object for each input sequence
        (more on this below). The ``VDJ`` objects should be appended to either the ``assigned`` or ``unassigned``
        attribute, depending on the outcome of the V(D)J assignment. A high-level overview of how ``__call__()``
        should work is as follows::

            def __call__(self, seqs, species):
                for seq in seqs:
                    vdj = self.assign_vdj(seq, species)
                    if self.is_properly_assigned(vdj):
                        self.assigned.append(vdj)
                    else:
                        self.unassigned.append(vdj)

        In this case, the output of ``assign_vdj()`` should be a ``VDJ`` object.


    V(D)J assignment
    ----------------

    AbStar defines ``VDJ`` objects, which are designed to package the input sequence and ``GermlineSegment``
    objects that provide information about germline assignment for V, D and/or J genes. ``VDJ`` objects must
    contain the input sequence and may also contain information about V, D and J assignments. While the
    sequence is required, the germline assignments are optional, to allow creation of ``VDJ`` objects for
    light chains (which lack a D-gene) and/or for unsuccessful assignments. Additionally, this allows
    flexibility in designing the ``__call__()`` method. Assigners can either instantiate a ``VDJ`` object at
    the start of ``__call__()`` and attach ``GermlineSegment``s as they are computed or they can instantiate
    ``VDJ`` objects upon completion of V(D)J assignment using the input sequences and computed
    ``GermlineSegment``s as instantiation arguments. Because ``VDJ`` objects contain built-in logging
    capability, such objects should be created for each input sequence even in the case of unsuccessful
    assignments so that assignment issues can be logged and reported.

    ``VDJ`` objects include a logging method: ``log()``. Using ``log()`` is similar to using Python's
    builtin ``print()`` function. Multiple arguments to ``log()`` will be space-delimited, and a ``sep``
    keyword allows you to provide an alternate delimiter. Log messages are typically used to record useful
    information or to note the occurance of errors/exceptions that are caught and handled by the Assigner.
    It is also a useful means by which to note the reason a specific input sequence is unable to be fully
    processed (because it isn't a valid antibody sequence, etc). By convention, AbStar log entries should
    contain the name of the log event (in all caps) followed by a colon and additional log information.
    An example ``log()`` call in the case of a V(D)J assignment that was aborted because the V-gene alignment
    score didn't meet the alignment quality threshold::

        for seq in seqs:
            vdj = VDJ(seq)
            v = assign_v_gene(seq)
            if v.score < threshold:
                vdj.log('V-GENE ASSIGNMENT ERROR:', 'Alignment score ({}) was too low'.format(v.score))
                continue

    ``VDJ`` objects also contain a mechanism for logging exceptions: ``exception()``. This is typically used
    to log unexpected exceptions (ie, those that are not caught and handled by the Assigner). It is useful
    to provide some information about when/where in the Assigner the exception occured, as well as a formatted
    version of the exception traceback. An example of ``exceptions()`` being used to log an exception that
    occurs during V-gene assignment::

        vdj = VDJ(seq)
        try:
            v = assign_v_gene(seq)
        except:
            vdj.exception('V-GENE ASSIGNMENT EXCEPTION:', traceback.format_exc())

    One of the primary benefits of recording log and exception information through the ``VDJ`` objects is
    that it allows for much easier formatting of such events when multiple sequences are being processed
    in parallel. Using Python's ``logging`` utility would result in log events being logged as they occur,
    which means that multiple log entries for the same input sequence may be far apart in the resulting log
    file. Packaging log and exception information with each ``VDJ`` object alows AbStar to create a much
    more useful log file in which all information related to a single input sequence is found in a single
    log location.


    ``GermlineSegment`` objects
    ---------------------------

    ``GermlineSegment`` objects contain information about the assigned germline gene segment (V, D or J).
    Instantiation of a ``GermlineSegment`` object requires only the name of the germline gene segment (in
    IMGT format, like 'IGHV3-23*01') and the species. Several other optional arguments can be included
    at instantiation. ``score`` is the assignment score, typically an ``int`` or ``float``. ``strand``
    indicates the strand orientation of the input sequence (``'+'`` or ``'-'``). ``others`` is a list
    of additional high scoring ``GermlineSegment`` objects. ``assigner_name`` is the name of the custom
    Assigner (as a ``str``), and will be converted to lowercase before recording in AbStar output.

    ``GermlineSegment`` objects also provide a few additional public attributes that may be desirable
    for certain custom Assigners, especially those that have been specifically designed to accurately
    define recombination regions. ``query_start``, ``query_end``, ``germline_start`` and ``germline_end``
    allow the Assigner to define the start and end points of the germline alignment for both the query
    and germline sequences. Note that both ``query_start`` and ``query_start`` and ``query_end`` MUST
    be relative to the complete query sequence in the ``'+'`` orientation (in other words, V-gene at
    the 5' end and J-gene at the 3' end). AbStar realigns all assigned germline genes using alignment
    parameters designed to accurately identify somatic mutations as well as somatic mutation-induced
    indels. If the additional assignment position attributes are provided by the Assigner, AbStar will
    force their use during realignment. If not, AbStar will use the assigned germline genes for realignment
    and will define the start and end points of the germline alignment during realignment.


    Example
    -------

    The following is an example of a custom Assigner class, named MyAssigner. It uses BioPython's SeqIO
    to parse the input file, and because BioPython is a dependency of AbStar, you can assume that BioPython
    will be present on any system running AbStar. All custom Assigners should be located in the ``assigners``
    directory::

        import os
        import platform
        import traceback

        from abstar.assigners.assigner import BaseAssigner
        from abstar.core.vdj import VDJ
        from abstar.core.germline import GermlineSegment

        from Bio import SeqIO

        class MyAssigner(BaseAssigner):

            def __init__(self, species):
                super(MyAssigner, self).__init__(species)
                self.binary = os.path.join(self.binary_directory, 'mybinary_{}'.format(platform.system()))

            def __call__(self, sequence_file, file_format):
                for sequence in SeqIO.parse(open(sequence_file, 'r'), file_format):
                    seq = Sequence(sequence)
                    vdj = VDJ(seq)

                    # V-gene assignment
                    # The VDJ object (rather than just the sequence) is passed to the assignment
                    # function for three reasons:
                    #    1) it contains the sequence, so we don't need to pass it separately
                    #    2) it allows logging during germline assignment, via vdj.log()
                    #    3) subsequent assignments will have access to information about
                    #       previous assignments (J-gene assignment operations have access
                    #       to information about the V-gene assignment, etc.)
                    try:
                        vdj.v = self.assign_germline(vdj, 'V')
                    except:
                        vdj.exception('V-ASSIGNMENT:', traceback.format_exc())

                    # J-gene assignment
                    try:
                        vdj.j = self.assign_germline(vdj, 'J')
                    except:
                        vdj.exception('J-ASSIGNMENT:', traceback.format_exc())

                    # D-gene assignment
                    try:
                        vdj.d = self.assign_germline(vdj, 'D')
                    except:
                        vdj.exception('D-ASSIGNMENT:', traceback.format_exc())

                    return vdj

            def assign_germline(self, vdj, segment):
                germ_db = os.path.join(self.germline_directory, '{}/ungapped/{}.fasta'.format(self.species,
                                                                                              segment.lower()))

                # do stuff to assign the germline gene, using the species-appropriate germline DB
                # ...
                # ...
                # ...
                # germs is a list of germline genes, ordered by score (highest first)

                if germs[0].score < threshold:
                    vdj.log('{}-ASSIGNMENT ERROR:'.format(segment),
                            'Score ({}) is too low'.format(germs[0].score))
                    return None
                others = [GermlineSegment(germ.name, self.species, score=germ.score) for germ in germs[1:6]]
                return GermlineSegment(germs[0].name, self.species, score=germs[0].score, others=others)

    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, species):
        super(BaseAssigner, self).__init__()
        self.name = self.__class__.__name__.lower()
        self.species = species
        self._assigned = None
        self._unassigned = None
        self._germline_directory = None
        self._binary_directory = None


    @abc.abstractmethod
    def __call__(self, sequence_file, file_format):
        pass


    @property
    def germline_directory(self):
        if self._germline_directory is None:
            self._germline_directory = get_germline_database_directory(self.species)
        return self._germline_directory

    @germline_directory.setter
    def germline_directory(self, directory):
        self._germline_directory = directory


    @property
    def binary_directory(self):
        if self._binary_directory is None:
            mod_dir = os.path.dirname(os.path.abspath(__file__))
            self._binary_directory = os.path.join(mod_dir, 'bin')
        return self._binary_directory

    @property
    def assigned(self):
        if self._assigned is None:
            self._assigned = []
        return self._assigned

    @assigned.setter
    def assigned(self, assigned):
        self._assigned = assigned

    @property
    def unassigned(self):
        if self._unassigned is None:
            self._unassigned = []
        return self._unassigned

    @unassigned.setter
    def unassigned(self, unassigned):
        self._unassigned = unassigned



#!/usr/bin/python
# filename: junction.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import math
import os
import traceback

from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from abtools import log
from abtools.alignment import global_alignment, local_alignment
from abtools.utils.codons import codon_lookup as codons


def get_junction(antibody):
    antibody.log('')
    antibody.log('JUNCTION')
    antibody.log('--------')
    try:
        return Junction(antibody)
    except Exception:
        antibody.exception('JUNCTION IDENTIFICATION ERROR', traceback.format_exc())
        raise


class Junction(object):
    """docstring for Junction"""
    def __init__(self, antibody):
        super(Junction, self).__init__()
        self.in_frame = True
        self.fallback_5prime = False
        self.fallback_3prime = False

        # identify junction sequence
        self.junction_nt_start = self._find_junction_nt_start(antibody)
        self.junction_nt_end = self._find_junction_nt_end(antibody)
        self.junction_nt = antibody.oriented_input[self.junction_nt_start:self.junction_nt_end]
        antibody.log('JUNCTION NT:', self.junction_nt, len(self.junction_nt))

        # check junction frame and retry junction identification with the slower
        # (but potentially more accurate) alignment method if it isn't in frame
        if len(self.junction_nt) % 3 != 0:
            antibody.log('WARNING: Junction out of frame!')
            if any([self.fallback_5prime is False, self.fallback_3prime is False]):
                # only try the 5' alignment if it wasn't already used
                if not self.fallback_5prime:
                    antibody.log('Attempting to identify junction start with fallback alignment method')
                    self.junction_nt_start = self._fallback_find_junc_nt_start(antibody)
                # only try the 3' alignment if it wasn't already used
                if not self.fallback_3prime:
                    antibody.log('Attempting to identify junction end with fallback alignment method')
                    self.junction_nt_end = self._fallback_find_junc_nt_end(antibody)
                self.junction_nt = antibody.oriented_input[self.junction_nt_start:self.junction_nt_end]
                antibody.log('JUNCTION NT:', self.junction_nt, len(self.junction_nt))
            if len(self.junction_nt) % 3 != 0:
                self.in_frame = False
        self.junction_aa = str(Seq(self.junction_nt.replace('-', ''), generic_dna).translate())
        antibody.log('JUNCTION AA:', self.junction_aa, len(self.junction_aa))

        # identify CDR3 sequence
        self.cdr3_aa = self.junction_aa[1:-1]
        self.cdr3_nt = self.junction_nt[3:-3]

        # parse N-addition regions
        if antibody.d is not None:
            n1_start = antibody.v.query_end + 1
            n1_end = antibody.d.query_start
            n2_start = antibody.d.query_end + 1
            n2_end = antibody.j.query_start
            self.v_nt = antibody.oriented_input[self.junction_nt_start:antibody.v.query_end + 1]
            self.n1_nt = antibody.oriented_input[n1_start:n1_end]
            self.d_nt = antibody.oriented_input[n1_end:n2_start]
            self.n2_nt = antibody.oriented_input[n2_start:n2_end]
            self.j_nt = antibody.oriented_input[n2_end:self.junction_nt_end]
            self.d_dist_from_cdr3_start_nt = antibody.d.query_start - self.junction_nt_start - 3
            self.d_dist_from_cdr3_end_nt = self.junction_nt_end - (antibody.d.query_end + 1) - 3
            self.n_nt = None
            antibody.log('V_NT:', self.v_nt)
            antibody.log('N1_NT:', self.n1_nt)
            antibody.log('D_NT:', self.d_nt)
            antibody.log('N2_NT:', self.n2_nt)
            antibody.log('J_NT:', self.j_nt)
            antibody.log('D-GENE DISTANCE FROM CDR3 START:', self.d_dist_from_cdr3_start_nt)
            antibody.log('D-GENE DISTANCE FROM CDR3 END:', self.d_dist_from_cdr3_end_nt)
        else:
            n_start = antibody.v.query_end + 1
            n_end = antibody.j.query_start
            self.v_nt = antibody.oriented_input[self.junction_nt_start:antibody.v.query_end + 1]
            self.n_nt = antibody.oriented_input[n_start:n_end]
            self.j_nt = antibody.oriented_input[n_end:self.junction_nt_end]
            self.n1_nt = None
            self.d_nt = None
            self.n1_nt = None
            self.d_dist_from_cdr3_start_nt = None
            self.d_dist_from_cdr3_end_nt = None
            antibody.log('V_NT:', self.v_nt)
            antibody.log('N_NT:', self.n_nt)
            antibody.log('J_NT:', self.j_nt)

        # calculate IMGT numbering for the junction and CDR3
        self.cdr3_imgt_nt_numbering = self._calculate_cdr3_imgt_nt_numbering()
        self.cdr3_imgt_aa_numbering = self._calculate_cdr3_imgt_aa_numbering()
        self.junction_imgt_nt_numbering = ['310', '311', '312'] + self.cdr3_imgt_nt_numbering + ['352', '353', '354']
        self.junction_imgt_aa_numbering = ['104'] + self.cdr3_imgt_aa_numbering + ['118']

        # if the J-gene is long enough to extend into the variable-length portion of the
        # junction numbering, need to adjust the IMGT numbering for the J-gene
        if min(antibody.j.imgt_aa_positions) < 112 or min(antibody.j.imgt_nt_positions) < 334:
            antibody.log('ADJUSTING J-GENE IMGT NUMBERING')
            antibody.log('OLD AA:', ', '.join([str(p) for p in antibody.j.imgt_aa_positions]))
            antibody.log('OLD NT:', ', '.join([str(p) for p in antibody.j.imgt_nt_positions]))
            self._adjust_jgene_imgt_numbering(antibody)
            antibody.log('NEW AA:', ', '.join([str(p) for p in antibody.j.imgt_aa_positions]))
            antibody.log('NEW NT:', ', '.join([str(p) for p in antibody.j.imgt_nt_positions]))
        else:
            antibody.j._correct_imgt_aa_position_from_imgt = {p: p for p in antibody.j.imgt_aa_positions}
            antibody.j._correct_imgt_nt_position_from_imgt = {p: p for p in antibody.j.imgt_nt_positions}


    def _find_junction_nt_start(self, antibody):
        start_codons = ['TGT', 'TGC']
        # position 310 is the start of codon 104, the conserved 2nd-Cys
        junc_start = antibody.v.get_raw_position_from_imgt(310)
        if junc_start is None:
            log_str = 'WARNING: The full 2nd-Cys codon does not appear to be present in the V-gene alignment. '
            log_str += 'Using fallback alignment method to find junction start.'
            antibody.log(log_str)
            return self._fallback_find_junc_nt_start(antibody)
        junc_start_codon = antibody.oriented_input[junc_start:junc_start + 3]
        antibody.log('JUNC START:', junc_start_codon, codons[junc_start_codon], junc_start)
        # if the identified junction start isn't normal, use the fallback (and slower)
        # alignment method for finding the junction start.
        if junc_start_codon not in start_codons:
            antibody.log('WARNING: Using fallback method to find junction start!')
            return self._fallback_find_junc_nt_start(antibody)
        return junc_start


    def _fallback_find_junc_nt_start(self, antibody):
        self.fallback_5prime = True
        # get the FR3 nt sequence of the IMGT gapped germline
        germ_fr3_sequence = antibody.v.imgt_germline.gapped_nt_sequence[196:309].replace('.', '')
        antibody.log('GERM FR3 SEQUENCE:', germ_fr3_sequence)
        # find the start of the junction (immediately after the end of FR3)
        aln = local_alignment(antibody.oriented_input, germ_fr3_sequence)
        antibody.log('  QUERY: ', aln.aligned_query)
        antibody.log('         ', aln.alignment_midline)
        antibody.log('GERM FR3:', aln.aligned_target)
        fr3_end = aln.query_end + (len(germ_fr3_sequence) - aln.target_end)
        junc_start_codon = antibody.oriented_input[fr3_end:fr3_end + 3]
        antibody.log('JUNC START:', junc_start_codon, codons[junc_start_codon], fr3_end)
        return fr3_end


    def _find_junction_nt_end(self, antibody):
        if antibody.chain == 'heavy':
            end_codons = ['TGG']
        else:
            end_codons = ['TTT', 'TTC']

        # when calculating the location of the conserved W/F, need to compensate
        # for sequences that don't contain the full J-gene
        joffset = len(antibody.j.raw_germline) - (antibody.j.germline_end + 1)
        # find the end of the J gene, then back up 12 codons to get to
        # the conserved W/F (start of FR4)
        junc_end = antibody.j.query_end - (33 - joffset) + 1
        junc_end_codon = antibody.oriented_input[junc_end:junc_end + 3]
        antibody.log('JUNC END:', junc_end_codon, codons[junc_end_codon], junc_end)

        # if the identified junction end isn't normal, use the fallback (and slower)
        # alignment method for finding the junction end.
        if junc_end_codon not in end_codons:
            antibody.log('WARNING: Did not identify conserved position 118. Using fallback method to find junction end.')
            return self._fallback_find_junc_nt_end(antibody)
        return junc_end + 3


    def _fallback_find_junc_nt_end(self, antibody):
        self.fallback_3prime = True

        # need to find the start of FR4 in the IMGT germline sequence
        end_res = 'W' if antibody.chain == 'heavy' else 'F'
        for i, res in enumerate(antibody.j.imgt_germline.ungapped_aa_sequence):
            if res == end_res and end_res not in antibody.j.imgt_germline.ungapped_aa_sequence[i + 1:]:
                fr4_nt_start_pos = (antibody.j.imgt_germline.coding_start - 1) + (i * 3)
                break
        germ_fr4_sequence = antibody.j.imgt_germline.gapped_nt_sequence[fr4_nt_start_pos:]

        # find the end of the junction (end of the first codon of FR4)
        aln = local_alignment(antibody.oriented_input, germ_fr4_sequence)
        fr4_start = aln.query_begin - aln.target_begin
        junc_end_codon = antibody.oriented_input[fr4_start:fr4_start + 3]
        antibody.log('JUNC END:', junc_end_codon, codons[junc_end_codon], fr4_start)
        return fr4_start + 3


    def _calculate_cdr3_imgt_nt_numbering(self):
        # if the length of the CDR3 is precisely 13, we don't need to adjust
        # the numbering at all (because the default IMGT numbering allows for exactly
        # 13 AA in the CDR3).
        if len(self.cdr3_aa) == 13:
            return list(range(313, 352))

        # If the CDR3 is shorter than 13 AA, we need to remove position numbers,
        # starting with positions to the left of position 333 (codon 111).
        elif len(self.cdr3_aa) < 13:
            gap = 13 - len(self.cdr3_aa)
            ltrim = int(math.ceil(float(gap) / 2)) * 3
            rtrim = int(math.floor(float(gap) / 2)) * 3
            lnums = list(range(313, 334 - ltrim))
            rnums = list(range(334 + rtrim, 352))
            return lnums + rnums

        # If the CDR3 is longer than 13 AA, we need to add position numbers (with decimal),
        # starting with positions to the right (position 334, or codon 112) and alternating
        # between left (position 333, codon 111) and right.
        # Nucleotides added to 334 will decrement from left to right, nucleotides added to 333
        # will increment from left to right. For example:
        # 331, 332, 333, 331.1, 331.2, 333.3, 334.3, 334.2, 334.1, 334, 335, ...
        # We add the nucleotide numbers in codon-length blocks, so the first codon will be
        # nucleotides 334.3, 334.2, and 334.1.
        elif len(self.cdr3_aa) > 13:
            bonus = len(self.cdr3_aa) - 13
            lbonus = int(math.floor(float(bonus) / 2)) * 3
            rbonus = int(math.ceil(float(bonus) / 2)) * 3
            lbonus_nums = ['333.{}'.format(i) for i in range(1, lbonus + 1)]
            rbonus_nums = ['334.{}'.format(i) for i in reversed(range(1, rbonus + 1))]
            return list(range(313, 334)) + lbonus_nums + rbonus_nums + list(range(334, 352))


    def _calculate_cdr3_imgt_aa_numbering(self):
        # Edge case for extremely short (likely non-productive) CDR3s
        if len(self.cdr3_aa) <= 2:
            return [105, 117][:len(self.cdr3_aa)]

        # if the length of the CDR3 is precisely 13, we don't need to adjust
        # the numbering at all (because the default IMGT numbering allows for exactly
        # 13 AA in the CDR3).
        if len(self.cdr3_aa) == 13:
            return list(range(105, 118))

        # If the CDR3 is shorter than 13 AA, we need to remove position numbers,
        # starting with positions to the left of codon 111 and then alternating left
        # and right.
        elif len(self.cdr3_aa) < 13:
            gap = 13 - len(self.cdr3_aa)
            ltrim = int(math.ceil(float(gap) / 2))
            rtrim = int(math.floor(float(gap) / 2))
            lnums = list(range(105, 112 - ltrim))
            rnums = list(range(112 + rtrim, 118))
            return lnums + rnums

        # If the CDR3 is longer than 13 AA, we need to add position numbers (with decimal),
        # starting with positions to the right (codon 112) and alternating
        # between left (codon 111) and right.
        # AAs added to 112 will decrement from left to right, AAs added to 111
        # will increment from left to right. For example:
        # 109, 110, 111, 111.1, 111.2, 112.3, 112.2, 112.1, 112, 113, ...
        elif len(self.cdr3_aa) > 13:
            bonus = len(self.cdr3_aa) - 13
            lbonus = int(math.floor(float(bonus) / 2))
            rbonus = int(math.ceil(float(bonus) / 2))
            lbonus_nums = ['111.{}'.format(i) for i in range(1, lbonus + 1)]
            rbonus_nums = ['112.{}'.format(i) for i in reversed(range(1, rbonus + 1))]
            return list(range(105, 112)) + lbonus_nums + rbonus_nums + list(range(112, 118))


    def _adjust_jgene_imgt_numbering(self, antibody):
        # adjust AA numbering
        aa_positions_to_replace = [p for p in antibody.j.imgt_aa_positions if p < 118]
        dot_positions = self.cdr3_imgt_aa_numbering[-len(aa_positions_to_replace):]
        new_positions = dot_positions + [p for p in antibody.j.imgt_aa_positions if p >= 118]
        antibody.j._correct_imgt_aa_position_from_imgt = {o: n for o, n in zip(antibody.j.imgt_aa_positions, new_positions)}
        antibody.j.imgt_aa_positions = new_positions
        # adjust NT numbering
        nt_positions_to_replace = [p for p in antibody.j.imgt_nt_positions if p < 352]
        dot_positions = self.cdr3_imgt_nt_numbering[-len(nt_positions_to_replace):]
        new_positions = dot_positions + [p for p in antibody.j.imgt_nt_positions if p >= 352]
        antibody.j._correct_imgt_nt_position_from_imgt = {o: n for o, n in zip(antibody.j.imgt_nt_positions, new_positions)}
        antibody.j.imgt_nt_positions = new_positions



#!/usr/bin/env python
# filename: germline.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import math
import os
import traceback

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from abtools.alignment import global_alignment, local_alignment
from abtools.sequence import Sequence
from abtools.utils.codons import codon_lookup
from abtools.utils.decorators import lazy_property

from ..utils.mixins import LoggingMixin


class GermlineSegment(object, LoggingMixin):
    """
    docstring for Germline

    Args:
    -----

        full (str): Full name of the germline gene, following IMGT naming conventions
            (eg: IGHV3-23*02).

        species (str): Species from which the germline gene was derived. Choices include
            'human', 'macaque', mouse', rabbit'.

        score (float): Score of the top germline alignment. Optional. If ``score`` is not
            provided, the realignment score will be used instead.

        strand (str): Strand of the alignment. Options are ``'+'`` for the positive strand
            (meaning the germline gene and the query sequence are in the same orientation)
            or ``'-'`` for the negative strand (query sequence is the reverse complement of
            the germline gene). Optional.

        others (list(Germline)): An optional list of additional high scoring germline genes.
            Can be a list of arbitrary length, with each member of the list being another
            ``Germline`` instance.

        assigner_name (str): The assigner name. Will be converted to lowercase. Optional.
            If not provided, ``assigner_name`` will be set to ``'unknown'``.
    """
    def __init__(self, full, species, score=None, strand=None, others=None, assigner_name=None):
        super(GermlineSegment, self).__init__()
        LoggingMixin.__init__(self)
        self.full = full
        self.species = species
        self.assigner_score = score
        self.strand = strand
        self.others = others
        self.gene_type = self.full[3].upper()
        self.assigner = assigner_name.lower() if assigner_name is not None else 'unknown'
        self._family = None
        self._gene = None
        self._chain = None

        # Optional properties for assigners to populate.
        # If populated by an assigner, they will be used to
        # force realignment to these parameters.
        #
        # Note that the query_start/query_end positions should
        # be 0-indexed and apply to the full query sequence. So if the sequence
        # is truncated following V-gene assignment and the truncated
        # sequence is submitted for J-gene assignment, make sure to
        # adjust the positions for the J-gene so that the numberings apply to
        # the complete input sequence, not the truncated sequence that was
        # used to identify the J-gene.
        #
        # Also note that because the end position of Python's slicing operator
        # is exclusive, to get the aligned sequence from the oriented_input, you
        # need to add one to the query_end (so that the slice includes query_end)
        self.query_start = None
        self.query_end = None
        self.germline_start = None
        self.germline_end = None

        # initialize log
        self.initialize_log()

        # These properties are populated by AbStar.
        # Assigners don't need to populate these (and they'll be overwritten
        # by AbStar if an assigner does populate them).
        self.score = None
        self.realignment = None
        self.raw_query = None
        self.raw_germline = None
        self.query_alignment = None
        self.germline_alignment = None
        self.alignment_midline = None
        self.alignment_length = None
        self.alignment_reading_frame = None  # 0-based, so if alignment start is the start of a codon, reading frame will be 0
        self.imgt_germline = None
        self.imgt_gapped_alignment = None
        self.imgt_nt_positions = []
        self.imgt_aa_positions = []
        self._imgt_position_from_raw = {}
        self._raw_position_from_imgt = {}
        self._correct_imgt_nt_position_from_imgt = None
        self._correct_imgt_aa_position_from_imgt = None
        self.fs_indel_adjustment = 0
        self.nfs_indel_adjustment = 0
        self.has_insertion = 'no'
        self.has_deletion = 'no'
        self._insertions = None
        self._deletions = None
        self.coding_region = None
        self.aa_sequence = None
        self.regions = None


    @property
    def family(self):
        if self._family is None:
            if '-' in self.full:
                self._family = self.full.split('-')[0]
        return self._family

    @family.setter
    def family(self, family):
        self._family = family


    @property
    def gene(self):
        if self._gene is None:
            if '*' in self.full:
                self._gene = self.full.split('*')[0]
        return self._gene

    @gene.setter
    def gene(self, gene):
        self._gene = gene


    @property
    def chain(self):
        if self._chain is None:
            c = {'H': 'heavy',
                 'K': 'kappa',
                 'L': 'lambda'}
            self._chain = c.get(self.full[2], None)
        return self._chain


    @property
    def insertions(self):
        if self._insertions is None:
            return []
        return self._insertions

    @insertions.setter
    def insertions(self, insertions):
        self._insertions = insertions


    @property
    def deletions(self):
        if self._deletions is None:
            return []
        return self._deletions

    @deletions.setter
    def deletions(self, deletions):
        self._deletions = deletions


    def correct_imgt_nt_position_from_imgt(self, position):
        if self._correct_imgt_nt_position_from_imgt is not None:
            return self._correct_imgt_nt_position_from_imgt.get(position, None)
        return None


    def correct_imgt_aa_position_from_imgt(self, position):
        if self._correct_imgt_aa_position_from_imgt is not None:
            return self._correct_imgt_aa_position_from_imgt.get(position, None)
        return None


    def initialize_log(self):
        log = []
        log.append('GERMLINE: {}'.format(self.full))
        self._log = log


    def realign_germline(self, antibody, query_start=None, query_end=None):
        '''
        Due to restrictions on the available scoring parameters in BLASTn, incorrect truncation
        of the v-gene alignment can occur. This function re-aligns the query sequence with
        the identified germline variable gene using more appropriate alignment parameters.

        Args:

            oriented_input (str): the raw input sequence, correctly oriented

            query_start (int): 5' position in `oriented_input` at which the sequence
                should be truncated prior to alignment with the germline sequence.

            query_end (int): 3' position in `oriented_input` at which the seqeunce
                should be truncated prior to alignment with the germline sequence
        '''
        oriented_input = antibody.oriented_input
        germline_seq = self._get_germline_sequence_for_realignment()
        aln_params = self._realignment_scoring_params(self.gene_type)
        # if the alignment start/end positions have been annotated by the assigner,
        # force re-alignment using those parameters
        if all([x is not None for x in [self.query_start,
                                        self.query_end,
                                        self.germline_start,
                                        self.germline_end]]):
            query = oriented_input.sequence[self.query_start:self.query_end]
            germline = germline_seq[self.germline_start:self.germline_end]
            alignment = global_alignment(query, germline, **aln_params)
        # use local alignment to determine alignment start/end positions if
        # they haven't already been determined by the assigner
        else:
            query = oriented_input.sequence[query_start:query_end]
            alignment = local_alignment(query, germline_seq, **aln_params)
        if alignment:
            self._process_realignment(antibody, alignment, query_start)
        else:
            antibody.log('GERMLINE REALIGNMENT ERROR')
            antibody.log('REALIGNMENT QUERY SEQUENCE:', query)
            antibody.log('QUERY START:', query_start)
            antibody.log('QUERN END:', query_end)


    def gapped_imgt_realignment(self):
        '''
        Aligns to gapped IMGT germline sequence. Used to determine
        IMGT-formatted position numberings so that identifying
        antibody regions is simplified.
        '''
        self.imgt_germline = get_imgt_germlines(species=self.species,
                                                gene_type=self.gene_type,
                                                gene=self.full)
        query = self.germline_alignment.replace('-', '')
        aln_params = self._realignment_scoring_params(self.gene_type)
        aln_params['gap_open'] = -11
        aln_matrix = self._get_gapped_imgt_substitution_matrix()
        self.imgt_gapped_alignment = local_alignment(query,
                                                     self.imgt_germline.gapped_nt_sequence,
                                                     matrix=aln_matrix,
                                                     **aln_params)
        self.alignment_reading_frame = ((2 * (self.imgt_gapped_alignment.target_begin % 3)) % 3) + (self.imgt_germline.coding_start - 1)  # IMGT coding start is 1-based
        self.coding_region = self._get_coding_region()
        self.aa_sequence = self._get_aa_sequence()
        try:
            self._imgt_numbering()
        except:
            self.exception('IMGT NUMBERING', traceback.format_exc(), sep='\n')


    def _get_gapped_imgt_substitution_matrix(self):
        matrix = {}
        residues = ['A', 'C', 'G', 'T', 'N', '.']
        for r1 in residues:
            matrix[r1] = {}
            for r2 in residues:
                if r1 == r2:
                    score = 3
                elif any([r1 == '.', r2 == '.']):
                    score = -3
                else:
                    score = -2
                matrix[r1][r2] = score
        return matrix

    def get_imgt_position_from_raw(self, raw):
        return self._imgt_position_from_raw.get(raw, None)


    def get_raw_position_from_imgt(self, imgt):
        return self._raw_position_from_imgt.get(imgt, None)


    def _process_realignment(self, antibody, aln, query_start):
        self.realignment = aln
        self.score = aln.score
        self.raw_query = aln.raw_query
        self.raw_germline = aln.raw_target
        self.query_alignment = aln.aligned_query
        self.germline_alignment = aln.aligned_target
        self.alignment_midline = ''.join(['|' if q == g else ' ' for q, g in zip(aln.aligned_query,
                                                                                 aln.aligned_target)])
        # only update alignment start/end positions if not already annotated by aligner
        if any([x is None for x in [self.query_start, self.query_end, self.germline_start, self.germline_end]]):
            offset = query_start if query_start is not None else 0
            self.query_start = aln.query_begin + offset
            self.query_end = aln.query_end + offset
            self.germline_start = aln.target_begin
            self.germline_end = aln.target_end
        self._fix_ambigs(antibody)
        self._find_indels(antibody)


    def _get_germline_sequence_for_realignment(self):
        '''
        Identifies the appropriate germline variable gene from a database of all
        germline variable genes.

        Returns:
        --------

            str: Germline sequence, or ``None`` if the requested germline gene could not be found.
        '''
        # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        # db_file = os.path.join(mod_dir, 'assigners/germline_dbs/{}_{}.fasta'.format(self.species.lower(), self.gene_type))
        germ_dir = get_germline_database_directory(self.species)
        db_file = os.path.join(germ_dir, 'ungapped/{}.fasta'.format(self.gene_type.lower()))
        try:
            for s in SeqIO.parse(open(db_file), 'fasta'):
                if s.id == self.full:
                    return str(s.seq)
            # TODO: log that the germline gene wasn't found in the database file
            return None
        except:
            # TODO: log that the germline database file couldn't be found
            return None


    def _imgt_numbering(self):
        aln_start = self.query_start
        aln_pos = 0
        imgt_start = self.imgt_gapped_alignment.target_begin + 1
        imgt_pos = imgt_start
        raw_position_from_imgt = {}
        imgt_position_from_raw = {}

        # imgt_start_offset is for J-genes only. Since the first position of the gapped IMGT
        # V-gene is the first position of the antibody seqeunce, IMGT numbering of the
        # V-gene should start at position 1. Not true with J-genes.
        # When processing V-genes, imgt_start_offset will always be 0.
        imgt_start_offset = self._get_imgt_start_offset()

        # Because we're iterating over the alignment but also want to track the raw (oriented_input)
        # query position, we need to adjust the alignment numbering in case there's a deletion
        # in the query sequence. With a deletion, the alignment position increases, but the raw
        # position shouldn't. query_del_adjustment allows us to adjust the alignment position so that
        # the raw query position is accurately tracked.
        query_del_adjustment = 0

        for gl in self.imgt_germline.gapped_nt_sequence[imgt_start:]:

            # only compute IMGT AA position numbers for V-genes. J-gene numbering is dependent on the CDR3 length
            # (due to IMGT's CDR3 naming conventions), so the junction must be annotated before J-gene AA position
            # numbers can be assigned.
            # if self.gene_type == 'V':

            # check to see if the position we're looking at is the start of a codon
            is_codon_start = (imgt_pos + imgt_start_offset) % 3 == 1

            # start by adding the aa position number (only if we're at the start of a codon)
            if is_codon_start:
                codon = self.germline_alignment[aln_pos:aln_pos + 3]
                if codon.count('-') >= 2:
                    self.imgt_aa_positions.append(None)
                else:
                    self.imgt_aa_positions.append((imgt_pos + imgt_start_offset + 2) / 3)

            # If the gapped IMGT germline is '.' (indicating a gap introduced by IMGT for numbering purposes),
            # we only need to increment the IMGT position and indicate the lack of sequence at the IMGT position.
            if gl == '.':
                raw_position_from_imgt[imgt_pos + imgt_start_offset] = None
                imgt_pos += 1
                continue

            # If there's a gap in the query alignment (deletion in the query sequence)
            # there's no equivalent IMGT position in the query.
            if self.query_alignment[aln_pos] == '-':
                raw_position_from_imgt[imgt_pos + imgt_start_offset] = None
                aln_pos += 1
                imgt_pos += 1
                query_del_adjustment += 1
                continue

            # If there's a gap in the germline alignment (insertion in the query sequence)
            # we need to increment the alignment position until the insertion is finished.
            # Since there isn't a germline IMGT position that's equivalent to the query position,
            # we don't want to increment the IMGT position or continue iterating through the IMGT
            # germline sequence.
            if self.germline_alignment[aln_pos] == '-':
                self.log('INFO: Found an insertion in the query sequence!')
                while self.germline_alignment[aln_pos] == '-':
                    imgt_position_from_raw[aln_pos + self.query_start - query_del_adjustment] = None
                    self.imgt_nt_positions.append(None)
                    aln_pos += 1

            # if the gapped IMGT germline isn't '.' and there's not an insertion in the query
            # sequence (or we've already iterated past it), we record both the IMGT position
            # and the raw (oriented_input) position and increment both the aligned and IMGt positions.
            raw_position_from_imgt[imgt_pos + imgt_start_offset] = aln_pos + aln_start - query_del_adjustment
            imgt_position_from_raw[aln_pos + aln_start - query_del_adjustment] = imgt_pos + imgt_start_offset
            self.imgt_nt_positions.append(imgt_pos + imgt_start_offset)
            aln_pos += 1
            imgt_pos += 1
            if aln_pos >= len(self.germline_alignment):
                break
        self._raw_position_from_imgt = raw_position_from_imgt
        self._imgt_position_from_raw = imgt_position_from_raw
        if self.insertions or self.deletions:
            self._calculate_imgt_indel_positions()


    def _get_imgt_start_offset(self):
        if self.gene_type == 'V':
            return 0
        # find the start of FR4 in the IMGT gapped germline gene
        end_res = 'W' if self.chain == 'heavy' else 'F'
        for i, res in enumerate(self.imgt_germline.ungapped_aa_sequence):
            if res == end_res and end_res not in self.imgt_germline.ungapped_aa_sequence[i + 1:]:
                nts_from_start_to_fr4 = (self.imgt_germline.coding_start) + (i * 3)
                break
        # the IMGT start offset is the conserved FR4 start position (352)
        # minus the number of nts from the start of the germline gene to FR4
        return 352 - nts_from_start_to_fr4


    def _get_coding_region(self):
        coding_region = self.query_alignment[self.alignment_reading_frame:].replace('-', '')
        truncation = len(coding_region) % 3
        if truncation > 0:
            coding_region = coding_region[:-truncation]
        return coding_region


    def _get_aa_sequence(self):
        return Seq(self.coding_region, generic_dna).translate()


    def _fix_ambigs(self, antibody):
        '''
        Fixes ambiguous nucleotides by replacing them with the germline nucleotide.
        '''
        self.query_alignment = ''.join([q if q.upper() != 'N' else g for q, g in zip(self.query_alignment,
                                                                                     self.germline_alignment)])
        # don't forget to also correct ambigs in the oriented_input sequence
        antibody.oriented_input.sequence = antibody.oriented_input.sequence[:self.query_start] + \
            self.query_alignment.replace('-', '') + \
            antibody.oriented_input.sequence[self.query_end + 1:]


    def _indel_check(self):
        if any(['-' in self.query_alignment, '-' in self.germline_alignment]):
            return True
        return False


    def _find_indels(self, antibody):
        '''
        Identifies and annotates indels in the query sequence.
        '''
        if self._indel_check():
            from ..utils import indels
            self.insertions = indels.find_insertions(antibody, self)
            if self.insertions:
                # only set self.has_insertion to 'yes' if the sequence has an in-frame insertion
                if 'yes' in [ins['in frame'] for ins in self.insertions]:
                    self.has_insertion = 'yes'
            self.deletions = indels.find_deletions(antibody, self)
            if self.deletions:
                # only set self.has_deletion to 'yes' if the sequence has an in-frame deletion
                if 'yes' in [deletion['in frame'] for deletion in self.deletions]:
                    self.has_deletion = 'yes'


    def _calculate_imgt_indel_positions(self):
        if self.insertions:
            for i in self.insertions:
                # since there's possibly not a direct IMGT correlate to the position
                # at the start of the insertion, need to find the closest IMGT correlate
                raw_pos = i.raw_position
                while self.get_imgt_position_from_raw(raw_pos) is None:
                    raw_pos -= 1
                i.imgt_position = self.get_imgt_position_from_raw(raw_pos)
                i.imgt_codon = int(math.ceil(i.imgt_position / 3.0))
        if self.deletions:
            for d in self.deletions:
                # since there's possibly not a direct IMGT correlate to the position
                # at the start of the deletion, need to find the closest IMGT correlate
                raw_pos = d.raw_position
                while self.get_imgt_position_from_raw(raw_pos) is None:
                    raw_pos -= 1
                d.imgt_position = self.get_imgt_position_from_raw(raw_pos)
                d.imgt_codon = int(math.ceil(d.imgt_position / 3.0))


    @staticmethod
    def _realignment_scoring_params(gene):
        '''
        Returns realignment scoring paramaters for a given gene type.

        Args:

            gene (str): the gene type ('V', 'D', or 'J')


        Returns:

            dict: realignment scoring parameters
        '''
        scores = {'V': {'match': 3,
                        'mismatch': -2,
                        'gap_open': -22,
                        'gap_extend': -1},
                  'D': {'match': 3,
                        'mismatch': -2,
                        'gap_open': -22,
                        'gap_extend': -1},
                  'J': {'match': 3,
                        'mismatch': -2,
                        'gap_open': -22,
                        'gap_extend': -1}}
        return scores[gene]


def get_germline_database_directory(species):
    addon_dir = os.path.expanduser('~/.abstar/germline_dbs')
    if os.path.isdir(addon_dir):
        if species.lower() in [os.path.basename(d[0]) for d in os.walk(addon_dir)]:
            return os.path.join(addon_dir, species.lower())
    mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    return os.path.join(mod_dir, 'assigners/germline_dbs/{}'.format(species.lower()))


def get_imgt_germlines(species, gene_type, gene=None):
    '''
    Returns one or more IMGTGermlineGene objects that each contain a single IMGT-gapped germline gene.

    Args:
    -----

        species (str): Species for which the germline genes should be obtained.

        gene_type (str): Options are 'V', 'D', and 'J'.

        gene (str): Full name of a germline gene (using IMGT-style names, like IGHV1-2*02).
                    If provided, a single ``IMGTGermlineGene`` object will be returned, or None if the
                    specified gene could not be found. If not provided, a list of ``IMGTGermlineGene``
                    objects for all germline genes matching the ``species`` and ``gene_type`` will be returned.

    Returns:
    --------

        IMGTGermlineGene: a single ``IMGTGermlineGene`` object (if ``gene`` is provided) or a list of
                          ``IMGTGermlineGene`` objects. If no sequences match the criteria or if a germline
                          database for the requested species is not found, ``None`` is returned.
    '''
    # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    # db_file = os.path.join(mod_dir, 'assigners/germline_dbs/imgt_gapped/{}_{}_imgt-gapped.fasta'.format(species, gene_type))
    germ_dir = get_germline_database_directory(species)
    db_file = os.path.join(germ_dir, 'imgt_gapped/{}.fasta'.format(gene_type.lower()))
    try:
        germs = [IMGTGermlineGene(g) for g in SeqIO.parse(open(db_file, 'r'), 'fasta')]
    except:
        # TODO: log that the germline database file couldn't be found

        # print('Could not locate the IMGT germline database file ({}).'.format(db_file))
        # print(traceback.format_exc())


        return None
    if gene is None:
        return germs
    try:
        return [g for g in germs if g.name == gene][0]
    except IndexError:


        # print('Could not locate the IMGT germline gene ({}).'.format(gene))
        # print(traceback.format_exc())


        return None


def get_germlines(species, gene_type, chain=None, gene=None):
    '''
    Returns one or more IMGTGermlineGene objects that each contain a single IMGT-gapped germline gene.

    Args:
    -----

        species (str): Species for which the germline genes should be obtained.

        gene_type (str): Options are 'V', 'D', and 'J'.

        chain (str): Options are 'heavy', 'kappa', and 'lambda'. If not provided, germline sequences
                     from all chains are returned.

        gene (str): Full name of a germline gene (using IMGT-style names, like IGHV1-2*02).
                    If provided, a single ``IMGTGermlineGene`` object will be returned, or None if the
                    specified gene could not be found. If not provided, a list of ``IMGTGermlineGene``
                    objects for all germline genes matching the ``species`` and ``gene_type`` will be returned.

    Returns:
    --------

        IMGTGermlineGene: a single ``IMGTGermlineGene`` object (if ``gene`` is provided) or a list of
                          ``IMGTGermlineGene`` objects. If no sequences match the criteria or if a germline
                          database for the requested species is not found, ``None`` is returned.
    '''
    germs = get_imgt_germlines(species, gene_type, gene=gene)
    if germs is None:
        return germs
    if chain is not None:
        c = chain[0].upper()
        germs = [g for g in germs if g.name[2] == c]
    if len(germs) == 1:
        return germs[0]
    return germs


class IMGTGermlineGene(object):
    """docstring for IMGTGermlineGene"""

    # species_lookup exists to translate IMGT names to AbStar germline DB names.
    # if the IMGT species isn't in species_lookup, AbStar will try to use the IMGT
    # species name directly.
    species_lookup = {'homo sapiens': 'human'}

    def __init__(self, sequence, species=None):
        self.raw_sequence = Sequence(str(sequence.seq), id=sequence.description)
        self._species = species
        self.gapped_nt_sequence = self.raw_sequence.sequence
        self.ungapped_nt_sequence = self.gapped_nt_sequence.replace('.', '')


    @lazy_property
    def accession(self):
        return self.raw_sequence.id.split('|')[0].strip()

    @lazy_property
    def name(self):
        return self.raw_sequence.id.split('|')[1].strip()

    @property
    def species(self):
        if self._species is None:
            species = self.raw_sequence.id.split('|')[2].strip().lower()
            self._species = self.species_lookup.get(species, species)
        return self._species

    @lazy_property
    def functionality(self):
        return self.raw_sequence.id.split('|')[3].strip()

    @lazy_property
    def gene_type(self):
        return self.raw_sequence.id.split('|')[4].strip()[0].upper()

    @lazy_property
    def coding_start(self):
        # NOTE: uses 1-based indexing, so need to adjust if using for slicing
        return int(self.raw_sequence.id.split('|')[7].strip())

    @lazy_property
    def nt_length(self):
        return int(self.raw_sequence.id.split('|')[12].strip().split('+')[0])

    @lazy_property
    def gap_length(self):
        return int(self.raw_sequence.id.split('|')[12].strip().split('+')[1].split('=')[0])

    @lazy_property
    def total_length(self):
        return int(self.raw_sequence.id.split('|')[12].strip().split('=')[1])

    @lazy_property
    def partial(self):
        p = []
        partial = self.raw_sequence.id.split('|')[13]
        if "3'" in partial:
            p.append("3'")
        if "5'" in partial:
            p.append("5'")
        return p

    @lazy_property
    def is_rev_comp(self):
        if self.raw_sequence.id.split('|')[14].strip() == 'rev-compl':
            is_rev_comp = True
        else:
            is_rev_comp = False
        return is_rev_comp

    @lazy_property
    def gapped_aa_sequence(self):
        res = []
        coding = self.gapped_nt_sequence[self.coding_start - 1:]
        for codon in (coding[pos:pos + 3] for pos in xrange(0, len(coding), 3)):
            if len(codon) != 3:
                continue
            if codon == '...':
                res.append('.')
            else:
                res.append(codon_lookup.get(codon, 'X'))
        return ''.join(res)

    @lazy_property
    def ungapped_aa_sequence(self):
        return self.gapped_aa_sequence.replace('.', '')
